#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Base options for querying Companies House REST API."""

import time
from dataclasses import dataclass, field
from datetime import datetime
from logging import getLogger
from math import ceil
from os import getenv
from typing import (
    AsyncGenerator,
    Dict,
    Final,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import requests
from dotenv import load_dotenv
from networkx import (
    Graph,
    compose,
    is_bipartite,
    is_connected,
    number_connected_components,
)
from requests.exceptions import ConnectionError
from requests.models import Response

from .company_codes import COMPANIES_HOUSE_URI_CODES
from .utils import (
    DEFAULT_API_KEY_PATH,
    Error,
    ExceededMaxBranchesException,
    InternetConnectionError,
    JSONDict,
    NegativeIntBranchException,
    QueryParameters,
    RunConfig,
    get_external_ip_address,
    get_kinds_ids_dict,
)

logger = getLogger(__name__)

load_dotenv(dotenv_path=DEFAULT_API_KEY_PATH)


COMPANIES_HOUSE_URL: Final[str] = "https://api.companieshouse.gov.uk"
COMPANIES_HOUSE_API_KEY_ENV_NAME: Final[str] = "COMPANIES_HOUSE_KEY"
COMPANIES_HOUSE_API_KEY: Final[Optional[str]] = getenv(
    COMPANIES_HOUSE_API_KEY_ENV_NAME
)

COMPANY_SUFFIXES: Final[Tuple[str, str, str]] = (
    "LTD",
    "LIMITED",
    "LLC",
)

COMPANIES_HOUSE_DATE_FORMAT: Final[str] = "%Y-%m-%d"

COMPANIES_HOUSE_TOTAL_RESULTS_KEYWORD: Final[str] = "total_results"
COMPANIES_HOUSE_ITEMS_QUERY_KEYWORD: Final[str] = "items_per_page"
COMPANIES_HOUSE_PAGINATION_KEYWORD: Final[str] = "start_index"
COMPANIES_HOUSE_APPOINTED_KEYWORD: Final[str] = "appointed_on"
COMPANIES_HOUSE_RESIGNATION_KEYWORD: Final[str] = "resigned_on"
COMPANIES_HOUSE_CEASED_KEYWORD: Final[str] = "ceased_on"

COMPANIES_HOUSE_QUERIES_COUNTS_KEYWORD: Final[str] = "items_per_query_list"

COMPANIES_HOUSE_DEFAULT_PAGINATION: Final[int] = 35
COMPANIES_HOUSE_MAX_OFFICER_PAGINATION: Final[int] = 50

CompanyIDType = str
CompanyRootIDsType = Union[CompanyIDType, Tuple[CompanyIDType, ...]]
JSONItemsGenerator = Generator[Tuple[str, JSONDict], None, None]

COMPANY_NETWORK_KINDS: Final[Tuple[str, str, str]] = (
    "company",
    "officer",
    "controller",
)

OFFICER_LINKS_KEY: Final[str] = "officers"
CONTROLLERS_LINKS_KEY: Final[
    str
] = "persons_with_significant_control_statements"


def companies_house_query(
    query: str,
    auth_key: Optional[str] = COMPANIES_HOUSE_API_KEY,
    sleep_time: int = 60,
    url_prefix: str = COMPANIES_HOUSE_URL,
    max_trials: int = 6,
    params: QueryParameters = None,
    items_per_page: int = COMPANIES_HOUSE_MAX_OFFICER_PAGINATION,
    items_per_page_keyword: str = COMPANIES_HOUSE_ITEMS_QUERY_KEYWORD,
    pagination: bool = True,
) -> Optional[JSONDict]:
    """Companies House API quiery repeated when necessary, returns json if valid.

    The `auth_tuple` reflects the `(username, password)` auth
    parameters customised for Companies House api which expects the `api_key`
    in the username spot and no separate component (hence blank password).

    Args:
        query (str): a query string such as '/company/04547069'
        auth_key (str): API key which is by default loaded from a .env file
        sleep_time (int): Number of seconds to pause after query error
        url_prefix (str): Prefix of url that defaults to COMPANIES HOUSE API
        max_trials (int): Number of attempts to query, default with sleep_time
                          exceeds the 5 min max query time
        params (dict): Dictionary of usually optional parameters for search

    Returns:
        dict: A Dict of a valid JSON response (`200` response code)
        None: None returned for response codes `404`, `500` and `502`

    Raises:
        Exception: If number of `trials` is exceeded.

    Todo:
        * Cover all exceptions in tests
        * Consider better means of managing 429 Too Many Requests error
          by checking header
    """
    auth_tuple: Tuple[Optional[str], str] = (auth_key, "")
    trials: int = max_trials
    params = params or {items_per_page_keyword: items_per_page}
    response: Response = requests.get(
        url_prefix + query,
        auth=auth_tuple,
        params=params,  # type: ignore[arg-type] # requests not hinted
    )
    while trials:
        try:
            response = requests.get(
                url_prefix + query,
                auth=auth_tuple,
                params=params,  # type: ignore[arg-type] # requests not hinted
            )
        except ConnectionError:
            raise InternetConnectionError
        if response.status_code == 200:
            if pagination:
                paginated_responses: List[Optional[JSONDict]] = [
                    response_dict
                    for response_dict in _expand_query_for_pagination(
                        query, response
                    )
                ]
                if paginated_responses and len(paginated_responses) > 1:
                    return _join_pagination_queries(paginated_responses)
                return paginated_responses[0]
                # import pdb; pdb.set_trace()

            # requery_count_required: Optional[int] = (
            #         expand_query_for_pagination(response.json()))
            # if requery_count_required:  # Int if required, Null if not
            #     params[COMPANIES_HOUSE_ITEMS_QUERY_KEYWORD] = (
            #             requery_count_required
            #             )
            #     return companies_house_query(query, auth_key, sleep_time,
            #                                  url_prefix, max_trials, params)
            # import pdb; pdb.set_trace()
            return response.json()
        logger.warning(f"Status code {response.status_code} from {query}")
        if response.status_code in (403, 401):
            raise CompaniesHousePermissionError(query)
        if response.status_code == 404:
            logger.error(f"Skipping {query}")
            return None
        if response.status_code == 500:
            logger.warning(f"Will skip after {max_trials - trials} repeat")
            if trials < max_trials - 1:
                return None
        if response.status_code == 502:
            # Server error, likely an overload issue (hence adding to wait)
            logger.warning(f"Adding a {sleep_time} sec wait")
            time.sleep(sleep_time)
        logger.warning(f"Trying again in {sleep_time} seconds...")
        time.sleep(sleep_time)
        trials -= 1
    raise Exception(
        f"Failed {max_trials} attempt(s) querying " f"{url_prefix + query}"
    )


def _expand_query_for_pagination(
    query: str,
    response: Response,
    params: QueryParameters = None,
    max_pagination: int = COMPANIES_HOUSE_MAX_OFFICER_PAGINATION,
    start_index_keyword: str = COMPANIES_HOUSE_PAGINATION_KEYWORD,
    total_results_keyword: str = COMPANIES_HOUSE_TOTAL_RESULTS_KEYWORD,
    current_per_page_keyword: str = COMPANIES_HOUSE_ITEMS_QUERY_KEYWORD,
    **kwargs,
) -> Generator[Optional[JSONDict], None, None]:
    """Return number of records if pagination exceeds results, None otherwise.

    Note:
        * Assumes the `total_results` component is applicable for all relevant
          cases.
    """
    params = params or {}
    response_dict: JSONDict = response.json()
    yield response_dict
    if (
        total_results_keyword in response_dict
        and current_per_page_keyword in response_dict
    ):
        current_per_page: int = response_dict[current_per_page_keyword]
        total_results: int = response_dict[total_results_keyword]
        per_page: int = max(current_per_page, max_pagination)
        params[current_per_page_keyword] = per_page
        if total_results > current_per_page:
            remaining_queries: int = ceil(
                (total_results - current_per_page) / max_pagination
            )
            for i in range(remaining_queries):
                params[start_index_keyword] = (i + 1) * per_page
                yield companies_house_query(
                    query, params=params, pagination=False, **kwargs
                )


def _join_pagination_queries(
    queries: List[Optional[JSONDict]],
    queries_count_keyword: str = COMPANIES_HOUSE_QUERIES_COUNTS_KEYWORD,
) -> JSONDict:
    """Join a paginated list of queries into a single dict.

    Todo:
        * Test if None portions of queries may be problematic (or can be
          eliminated)
    """
    # joined_dict: dict = queries[0]
    # joined_dict[queries_count_keyword] = [
    #     len(joined_dict["items"]),
    # ]
    # for i, json_dict in enumerate(queries[1:]):
    # joined_dict: TypedDict("PaginatedQueries", {"")
    joined_dict: JSONDict = {queries_count_keyword: [], "items": []}
    for i, json_dict in enumerate(queries):
        if json_dict:
            try:
                joined_dict[queries_count_keyword].append(
                    len(json_dict["items"])
                )
                joined_dict["items"].extend(json_dict["items"])
            except TypeError as err:
                logger.warning(
                    f"Could not extend dict of pagination records for "
                    f"{joined_dict['links']['self']}\n"
                    f"Error: '{err}'\n"
                    # i+2 to account for queries[1:] + 1 for 0 indexing
                    # f"at {i+2} of {len(queries)} queries."
                    # Changed to relect shift of iteraction
                    f"at {i+1} of {len(queries)} queries."
                )
    return joined_dict


class CompaniesHousePermissionError(Error):
    """An exception for handling 403 forbidden errors."""

    def __init__(
        self,
        query: str = None,
        message: str = None,
    ) -> None:
        """Either set passed message or set to _default_error_message()."""
        self.query = query
        self.message = message or self._default_error_message()

    def __str__(self) -> str:
        """Return error message."""
        return self.message

    def _default_error_message(self) -> str:
        """Try to check current IP address to raise clear permission error."""
        ip_address = get_external_ip_address()
        return (
            f"Query: {self.query}\nreturned a 403 (forbidden) error. "
            "If that query seems correct, check the "
            f"{COMPANIES_HOUSE_API_KEY_ENV_NAME} is set in your local .env "
            "file.\n"
            "If both are correct, check the external IP address of "
            f"this computer ({ip_address}) is included in the list of "
            "Restricted IPs on your registered Companies House API Key."
        )


def stringify_company_id(company_id: Union[int, str]) -> str:
    """Enforce correct company number as string of length >= 8."""
    company_id = str(company_id)
    if company_id and len(company_id) < 8:
        return company_id.rjust(8, "0")  # Shorter need preceeding '0's
    return company_id


@dataclass
class CompanyRunConfig(RunConfig):
    """RunConfig dict with added root_id."""

    root_id: Optional[CompanyRootIDsType] = None
    kind: Final[str] = "company"
    composed_runs: List["CompanyRunConfig"] = field(default_factory=list)


class CompanyNetworkClient:
    """Recursively construct a network of companies and board members.

    Todo:
        * Consider abstracting to generically follow links
        * Look up 'persons_with_significant_control_statements':
          '/company/04547069/persons-with-significant-control-statements'
          (a PUNCHDRUNK link)
    """

    def __init__(
        self,
        branches: int = 0,
        include_significant_controllers: bool = False,
        include_officers: bool = True,
        include_edge_data: bool = False,
        enforce_missing_ties: bool = False,
        exclude_non_active_companies: bool = False,
        exclude_resigned_board_members: bool = False,
        exclude_ceased_controllers: bool = False,
        reset_cache: bool = True,
        compose_queried_networks: bool = False,
    ) -> None:
        """Initialise elements for query parameters."""
        if type(branches) is int and branches >= 0:
            self.branches = branches
        else:
            raise NegativeIntBranchException(branches)
        self.include_significant_controllers = include_significant_controllers
        self.include_officers = include_officers
        self.include_edge_data = include_edge_data
        self.enforce_missing_ties = enforce_missing_ties
        self.exclude_non_active_companies = exclude_non_active_companies
        self.exclude_resigned_board_members = exclude_resigned_board_members
        self.exclude_ceased_controllers = exclude_ceased_controllers
        self.compose_queried_networks = compose_queried_networks

        # Initiate `` _reset_cache`` to support
        # application of ``compose_queries_required`` but worth leaving
        # independet in some cases
        self._reset_cache = reset_cache
        self._root_id: Optional[CompanyIDType] = None
        self._graph: Graph = Graph()
        self._officer_appointments_cache: Dict = {}
        self._runs: List[CompanyRunConfig] = []

    @property
    def reset_cache(self) -> bool:
        """Return current ``reset_cache`` state."""
        return self._reset_cache

    @reset_cache.setter
    def reset_cache(self, value: bool) -> None:
        """Set ``self._reset_cache`` to value."""
        self._reset_cache = value

    @property
    def _parameter_state(self) -> QueryParameters:
        """Return the state of the parameters passable to __init__."""
        return {
            varname: getattr(self, varname)
            for varname in self.__init__.__code__.co_varnames  # type: ignore
            if varname != "self"
        }

    @_parameter_state.setter
    def _parameter_state(self, var: QueryParameters) -> None:
        for varname, value in var.items():
            setattr(self, varname, value)

    @property
    def _root_node_ids(
        self,
    ) -> Generator[Optional[CompanyRootIDsType], None, None]:
        """Return a tuple of root nodes from previous queries."""
        for config in self._runs:
            yield config.root_id

    def get_network(self, root_id: CompanyIDType = "04547069") -> Graph:
        """Construct a board interlock network using the NetworkX library.

        Todo:
            * Consider adding logs for each run
            * Refactor considering different use cases
        """
        root_id = stringify_company_id(root_id)
        if (
            self.reset_cache
            or not len(self._graph)
            or self.compose_queried_networks
        ):
            self._runs.append(
                CompanyRunConfig(
                    root_id=root_id,
                    start_time=None,
                    # logs=None,
                    # kind="company",
                    # kinds_ids_dict=None,
                    end_time=None,
                    parameter_state=self._parameter_state,
                )
            )
            if len(self._runs) > 1 and (
                self._runs[-1].parameter_state
                != self._runs[-2].parameter_state
            ):
                logger.warning(
                    "Query parameters differ between "
                    f"{self._runs[-1]} "
                    f"and {self._runs[-1]}"
                )
            # It is crucial to change here rather than pass to _get_network()
            self._root_id = root_id
            if not self.compose_queried_networks or self.reset_cache:
                self._graph = Graph()
                self._officer_appointments_cache = {}
            self._get_network()
            self._runs[-1].kinds_ids_dict = get_kinds_ids_dict(
                self._graph, COMPANY_NETWORK_KINDS
            )
            self._runs[-1].start_time = self._query_start
            self._runs[-1].end_time = self._query_end
            self._runs[
                -1
            ].connected_components_count = number_connected_components(
                self._graph
            )

        return self._graph

    def networks_generator(
        self,
        root_ids: Iterable[CompanyIDType],
        parameter_states: Iterator[QueryParameters] = None,
        composed_query: bool = False,
    ) -> Generator[Sequence[Graph], None, None]:
        """Query the ``root_ids`` list in order."""
        for root_id in root_ids:
            if composed_query:
                logger.info(f"Composed query of company {root_id}")
            if parameter_states:
                self._parameter_state = next(parameter_states)
            yield self.get_network(root_id)

    def get_composed_network(self, *args, **kwargs):
        """Iterate over querires then return self._graph."""
        [
            g
            for g in self.networks_generator(
                composed_query=True, *args, **kwargs
            )
        ]
        return self._graph

    async def async_networks_generator(
        self,
        root_ids: Iterable[CompanyIDType],
        parameter_states: Iterator[QueryParameters] = None,
        composed_query: bool = False,
    ) -> AsyncGenerator[Sequence[Graph], None]:
        """Async query the ``root_ids`` list in order.

        Todo:
            * Implement an async iterable for root_ids
        """
        # async for root_id in root_ids:
        for root_id in root_ids:
            if composed_query:
                logger.info(f"Composed async query of company {root_id}")
            if parameter_states:
                self._parameter_state = next(parameter_states)
            yield self.get_network(root_id)

    async def async_get_composed_network(self, *args, **kwargs) -> Graph:
        """Async iterate over querires then return self._graph."""
        [
            g
            async for g in self.async_networks_generator(
                composed_query=True, *args, **kwargs  # type: ignore
            )
        ]
        return self._graph

    def _get_officer_name(
        self,
        officer_id: str,
        company_id: str = None,
        board_data: JSONDict = None,
    ) -> str:
        """Get officer name from appointments list or company relationship.

        If a `company_id` is passed then attempt to return the name entry in
        the `_officer_appointments_cache`. If a KeyError, try to return the
        name listed in the relative companies officers list.

        Finally: if no company_id is provided then return the first name listed
        in contracts in the `_officer_appointments_cache`.

        Todo:
            * Add a method to take advantage of `"name_elements"`
            * Test the various exceptions/logs
            * Test the `return board_data['name']` option
            * Fix the appointment list and perhaps add log for that
            * Consider refactoring for efficiency
        """
        if company_id and board_data:
            try:
                name = self._officer_appointments_cache[officer_id][
                    company_id
                ]["name"]
                if name:
                    return name
            except KeyError:
                logger.warning(
                    "No 'name' data available for "
                    f'officer {board_data["name"]} '
                    f"({officer_id}) in appointments_cache"
                )
            try:
                name = board_data["name"]
            except KeyError:
                logger.warning(
                    "No 'name' data available for "
                    f"officer {officer_id} "
                    f"for company {company_id}"
                )
                return ""
            if not name:  # Check name isn't null
                logger.warning(
                    "Null name listed for "
                    f"officer {officer_id} "
                    f"from company {company_id}"
                )
            return name
        else:
            for appointment in self._graph.nodes[officer_id]["data"]["items"]:
                return appointment["name"]
            return ""

    def _include_officers(
        self, company_id: str, company_info_dict: dict, branch_iterator: int
    ) -> None:
        officers_data = get_company_officers_data(company_id)
        company_info_dict["officers"] = officers_data
        officers = get_company_officers(
            company_id, self.exclude_resigned_board_members, officers_data
        )
        for officer_id, officer_board_data in officers:
            self._add_person_edge_branch(
                officer_id, company_id, officer_board_data, branch_iterator
            )

    def _include_controllers(
        self, company_id: str, company_info_dict: dict, branch_iterator: int
    ) -> None:
        controllers_data = get_significant_controllers_data(company_id)
        company_info_dict["significant_controllers"] = controllers_data
        controllers = get_significant_controllers(
            company_id, controllers_data, self.exclude_ceased_controllers
        )
        for controller_id, controller_company_data in controllers:
            self._add_person_edge_branch(
                controller_id,
                company_id,
                controller_company_data,
                branch_iterator,
                method_name="_add_controller",
            )

    def _add_officer(
        self, officer_id: str, company_id: str, officer_board_data: JSONDict
    ) -> None:
        appointments_data = get_officer_appointments_data(officer_id)
        if appointments_data and "items" in appointments_data:
            if self.exclude_resigned_board_members:
                self._officer_appointments_cache[officer_id] = {
                    appointment["appointed_to"]["company_number"]: appointment
                    for appointment in appointments_data["items"]
                    if COMPANIES_HOUSE_RESIGNATION_KEYWORD not in appointment
                }
                resigend_board_positions_count: int = len(
                    [
                        appointment
                        for appointment in appointments_data["items"]
                        if COMPANIES_HOUSE_RESIGNATION_KEYWORD in appointment
                    ]
                )
                logger.debug(
                    f"Skipping {resigend_board_positions_count} "
                    f"resigned board positions for officer "
                    f"{officer_id}"
                )
            else:
                self._officer_appointments_cache[officer_id] = {
                    appointment["appointed_to"]["company_number"]: appointment
                    for appointment in appointments_data["items"]
                }
        name = self._get_officer_name(
            officer_id, company_id, officer_board_data
        )
        self._graph.add_node(
            officer_id,
            name=name,
            bipartite=1,
            kind=COMPANY_NETWORK_KINDS[1],
            is_person=is_person(name),
            data=appointments_data,
        )

    def _add_controller(
        self,
        controller_id: str,
        company_id: str,
        controller_company_data: JSONDict,
    ) -> None:
        """Add a significant_controller node to self._graph.

        Todo:
            * See if there are possibilities for foreign keys to be duplicated
            * Take advantage of different company/individual queries/types
            * Consider a cache option if IDs prove generally unique
        """
        controller_individual_data = (
            get_significant_controller_person_or_company_data(
                controller_company_data
            )
        )
        name = controller_company_data["name"]
        is_individual = is_individual_controller_url(
            controller_company_data["links"]["self"]
        ) & is_person(name)
        self._graph.add_node(
            controller_id,
            name=name,
            bipartite=1,
            kind=COMPANY_NETWORK_KINDS[2],
            is_person=is_individual,
            data=controller_individual_data,
        )

    def _add_person_edge_branch(
        self,
        person_id: str,
        company_id: str,
        data: JSONDict,
        branch_iterator: int,
        method_name="_add_officer",
    ) -> None:
        if person_id not in self._graph.nodes:
            getattr(self, method_name)(person_id, company_id, data)
        self._graph.add_edge(company_id, person_id, data=data)
        if 0 <= branch_iterator < self.branches:
            self._get_network_branches(person_id, branch_iterator)
        elif branch_iterator < 0:
            raise NegativeIntBranchException(branch_iterator)
        elif branch_iterator > self.branches:
            raise ExceededMaxBranchesException(branch_iterator, self.branches)

    def _get_network(
        self, company_id: Optional[str] = None, branch_iterator: int = 0
    ) -> None:
        company_id = company_id or self._root_id
        assert company_id
        if company_id == self._root_id:
            self._query_start = datetime.now()
        logger.debug(f"Querying board network from {company_id}")
        company_data = get_company_data(
            company_id,
            exclude_non_active_companies=self.exclude_non_active_companies,
        )
        if not company_data:
            return None
        logger.debug(company_data["company_name"])
        company_info_dict = {"company": company_data}
        self._graph.add_node(
            company_id,
            name=company_data["company_name"],
            kind="company",
            is_person=False,
            category=get_company_category(company_id),
            bipartite=0,
            data=company_info_dict,
        )
        if (
            self.include_officers
            and "links" in company_data
            and OFFICER_LINKS_KEY in company_data["links"]
        ):
            self._include_officers(
                company_id, company_info_dict, branch_iterator=branch_iterator
            )
        if self.include_significant_controllers:
            self._include_controllers(
                company_id, company_info_dict, branch_iterator=branch_iterator
            )
        if company_id == self._root_id:
            self._query_end = datetime.now()

    def _get_network_branches(
        self,
        person_id: str,
        branch_iteration: int,
        cache_name: str = "_officer_appointments_cache",
    ):
        """Query further network branches from board members.

        Todo:
            * Explore further branch potentials via controllers.
            * Assess if it's better to choose whether to get more branches here
              or in self._add_person_edge_branch
        """
        if not hasattr(self, cache_name):
            raise KeyError(f"{cache_name} not set so cannot add that branch")
        try:
            for related_company_id in getattr(self, cache_name)[person_id]:
                if related_company_id not in self._graph.nodes:
                    self._get_network(related_company_id, branch_iteration + 1)
                    if self.enforce_missing_ties:
                        logger.warning(
                            f"Enforcing possible tie between "
                            f"{person_id} and {related_company_id}"
                        )
                        self._graph.add_edge(
                            person_id, related_company_id, data=None
                        )
        except KeyError:
            logger.warning(f"No branch data from {person_id}")


def get_company_network(
    company_id: CompanyIDType = "04547069",
    branches: int = 0,
    include_significant_controllers: bool = False,
    include_officers: bool = True,
    include_edge_data: bool = False,
    **kwargs,
) -> Optional[Graph]:
    """Recursively query a bipartite network of companies and boards.

    This function returns a bipartite NetworkX `Graph` where bipartite is 0 for
    companies and 1 for board members.

    Args:
        company_id (CompanyIDType): An int or str of a company_id
        branches (int): A positive number of hops to follow in snowball
            sampling from board members to other companies they set on
        include_significant_controllers (bool): Whether to also follow
            significant controllers as another type of board member
        include_officers (bool): Include officers in network queries.
        include_edge_data (bool): Include JSON data from
            `get_officer_appointments` as a `data` attribute for edges.
        params (dict): Dictionary of usually optional parameters for search
        **kwargs: Additional parameters to send to api calls.

            exclude_non_active_companies (bool) is for `get_company`.

            exclude_resigned_board_members (bool) is for
            `get_company_officers`.

            enforce_missing_ties (bool) is for `_get_network_branches`.

    Returns:
        Graph: A bipartite NetworkX `Graph` where bipartite is 0 for
               companies and 1 for board members.
        None: If no company is found.

    Todo:
        * 429 Too Many Requests error raised if > 600/min
        * Test officers error on company '01086582'
        * Consider removing print statement within the related loop
        * Refactor todo info into documentation
        * Add option of including network collected prior to error
        * Add options for include_significant_controllers
        * Consider enforcing inclusion of data in edges
        * Consider means of caching queries to avoid duplicates
    """
    g = Graph()
    logger.debug(f"Querying board network from {company_id}")
    company = get_company_data(company_id, **kwargs)
    if not company:
        return None
    logger.debug(company["company_name"])
    g.add_node(
        company_id,
        name=company["company_name"],
        category=get_company_category(company_id),
        bipartite=0,
        data=company,
    )
    if include_officers and (
        "links" in company and OFFICER_LINKS_KEY in company["links"]
    ):
        officers = get_company_officers(company_id, **kwargs)
        for officer_id, officer_data in officers:
            g.add_node(
                officer_id,
                name=officer_data["name"],
                bipartite=1,
                data=officer_data,
            )
            if branches or include_edge_data:
                # fmt: off
                appointments = {
                    related_company_id: appointment_data
                    for related_company_id, appointment_data in
                    get_officer_appointments(
                        officer_id,
                        company=company,
                        company_id=company_id,
                        officer_data=officer_data,
                    )
                }
                # fmt: on
            if include_edge_data:
                # consider poping key from appointments to avoid excess
                # loop
                try:
                    appointment_data: Optional[JSONDict] = appointments[
                        company_id
                    ]
                except KeyError:
                    logger.warning(
                        "No appointment_data available for "
                        f'officer {officer_data["name"]} '
                        f"({officer_id})"
                    )
                    appointment_data = None
                officer_edge_data = {
                    "appointment_data": appointment_data,
                    # 'officer_data is a duplicate of officer node data
                    "officer_data": officer_data,
                }
                g.add_edge(company_id, officer_id, data=officer_edge_data)
            else:
                g.add_edge(company_id, officer_id)
            if branches:
                g = _get_network_branches(
                    g,
                    officer_id,
                    appointments,
                    branches=branches,
                    root_id=company_id,
                    root_company_data=company,
                    **kwargs,
                )
    return g


def _get_network_branches(
    g: Graph,
    officer_id: str,
    appointments: Dict[str, JSONDict],
    root_id: CompanyIDType,
    root_company_data: JSONDict,
    # Argument ordering hopefully refactorable
    branches: int = 0,
    enforce_missing_ties: bool = False,
    include_edge_data: bool = True,
    **kwargs,
) -> Optional[Graph]:
    """Recursively expand network through individuals on multiple boards.

    Args:
        g (Graph): A graph meant to have at least one company.
        appointments (dict): Dictionary of company_id keys to appointment data.
        branches (int): Number of branches to follow.
        enforce_missing_ties (bool): Whether to add ties in cases where there
            is a record of a board membership in `get_officer_appointments`
            that doesn't appear in that company's `get_company_officers` query.
        include_edge_data (bool): Include JSON data from
            `get_officer_appointments` as a `data` attribute for edges.
        root_id (CompanyIDType): int or str of root `company_id` branch
            may connect to
        root_company_data (JSONDict): A dict of data on company branch may
            connect to
        **kwargs: Parameters to send to further calls of `get_company_network`.

    Returns:
        Graph: A bipartite NetworkX `Graph` where bipartite is 0 for
               companies and 1 for board members.

    Todo:
        * Expect abstract refactoring to also work with
          `significant_controllers`
        * Consider refactor to always include edge data
        * Consider means of caching queries to avoid duplicates
    """
    for related_company_id in appointments:
        if related_company_id not in g.nodes:
            related_network = get_company_network(
                related_company_id,
                branches=branches - 1,
                enforce_missing_ties=enforce_missing_ties,
                include_edge_data=include_edge_data,
                **kwargs,
            )
            if related_network:
                g = compose(g, related_network)
                assert is_bipartite(g)
                if not is_connected(g) and enforce_missing_ties:
                    if include_edge_data:
                        g.add_edge(
                            related_company_id,
                            officer_id,
                            data=appointments[related_company_id],
                        )
                    else:
                        g.add_edge(related_company_id, officer_id)
            else:
                logger.warning(
                    "Skipping company "
                    f"{related_company_id} "
                    "from board member "
                    f"{g.nodes['officer_id']['name']} "
                    f"({officer_id}) of company "
                    f"{root_company_data['company_name']} "
                    f"({root_id})"
                )
    return g


def get_company_data(
    company_id: CompanyIDType = "04547069",
    exclude_non_active_companies: bool = False,
    **kwargs,
) -> Optional[JSONDict]:
    """Query company data."""
    company_id = stringify_company_id(company_id)
    company = companies_house_query("/company/" + company_id)
    if not company:
        logger.error(f"Querying data on company {company_id} failed")
        return None
    if exclude_non_active_companies:
        if (
            "company_status" in company
            and company["company_status"] != "active"
        ):
            logger.warning(
                f"Excluding company {company_id} because "
                f'status is {company["company_status"]}. '
                f'Company name: {company["company_name"]}'
            )
            return None
    return company


def get_company_officers_data(
    company_id: CompanyIDType = "04547069",
) -> Optional[JSONDict]:
    """Query company officer data."""
    officers_query = companies_house_query(f"/company/{company_id}/officers")
    if not officers_query:
        logger.error(f"Error requesting officers of company {company_id}")
        # Worth considering saving error here
    return officers_query


def get_company_officers(
    company_id: CompanyIDType = "04547069",
    exclude_resigned_board_members: bool = False,
    officers_query: JSONDict = None,
    **kwargs,
) -> JSONItemsGenerator:
    """Yield officer_id and officer data from company_id's board.

    Todo:
        * Refactor to return whole query to company for meta-data
    """
    if not officers_query:
        officers_query = get_company_officers_data(company_id)
    if officers_query:
        for officer in officers_query["items"]:
            if exclude_resigned_board_members:
                if is_inactive(officer):
                    logger.debug(
                        f"Skipping officer {officer['name']} because "
                        f"of resignation on {officer['resigned_on']}"
                    )
                    continue
            officer_id = officer["links"]["officer"]["appointments"].split(
                "/"
            )[2]
            logger.debug(f'{company_id} {officer["name"]} {officer_id}')
            yield officer_id, officer


def get_officer_appointments(
    officer_id: str = None, appointments: JSONDict = None, **kwargs
) -> JSONItemsGenerator:
    """Query officer appointments and yield company_id, appointment_data."""
    if officer_id and not appointments:
        appointments = get_officer_appointments_data(officer_id, **kwargs)
    if appointments:
        for appointment in appointments["items"]:
            yield appointment["appointed_to"]["company_number"], appointment


def get_officer_appointments_data(
    officer_id: str, **kwargs
) -> Optional[JSONDict]:
    """Query raw officer appointments data and return as JSON if possible."""
    appointments = companies_house_query(
        f"/officers/{officer_id}/appointments"
    )
    if not appointments:
        # Worth considering saving exact error
        if {"company", "company_id", "officer_data"} <= kwargs.keys():
            company, company_id, officer_data = (
                kwargs["company"],
                kwargs["company_id"],
                kwargs["officer_data"],
            )
            logger.error(
                "Error requesting appointments of board "
                f"member {officer_data['name']} ({officer_id}) of "
                f"company {company['company_name']} "
                f"({company_id})"
            )
        else:
            logger.error(
                "Error requesting appointments of board "
                f"member {officer_id}"
            )
    return appointments


def get_significant_controllers_data(
    company_id: str, **kwargs
) -> Optional[JSONDict]:
    """Return full JSON from a significant_controllers query."""
    persons = companies_house_query(
        f"/company/{company_id}/persons-with-significant-control"
    )
    if not persons:
        logger.error(
            "Error requesting significant controllers from company "
            f"{company_id}"
        )
    return persons


def get_significant_controllers(
    company_id: str,
    controllers_data: Optional[JSONDict] = None,
    exclude_ceased_controllers: bool = False,
    **kwargs,
) -> JSONItemsGenerator:
    """Return a generator of significant controller json records."""
    if not controllers_data:
        controllers_data = get_significant_controllers_data(
            company_id, **kwargs
        )
    if controllers_data:
        for controller in controllers_data["items"]:
            if exclude_ceased_controllers:
                if is_inactive(
                    controller, keyword=COMPANIES_HOUSE_CEASED_KEYWORD
                ):
                    controllers_ceased_date = controller[
                        COMPANIES_HOUSE_CEASED_KEYWORD
                    ]
                    logger.debug(
                        f"Skipping controller "
                        f"{controller['name']} because they "
                        f"ceased on {controllers_ceased_date}"
                    )
                continue
            controller_id = controller["links"]["self"].split("/")[-1]
            logger.debug(f'{company_id} {controller["name"]} {controller_id}')
            yield controller_id, controller


def get_significant_controller_person_or_company_data(
    controller_data: JSONDict,
) -> Optional[JSONDict]:
    """Query individual data on a significant_controller."""
    data: Optional[JSONDict]
    link: str = controller_data["links"]["self"]
    link_list: List[str] = link.split("/")
    if link_list[-2] == "individual":
        data = get_significant_controller_person_data(
            link_list[-4], link_list[-1]
        )
    elif link_list[-4] == "corporate-entity":
        data = get_significant_controller_company_data(
            link_list[-4], link_list[-1]
        )
    else:
        logger.warning(
            "Querying an unsupported significant controller type" f"{link})"
        )
        data = companies_house_query(link)
    if not data:
        logger.error(
            f"Error requesting data on significant controller from " f"{link}"
        )
    return data


def get_significant_controller_person_data(
    company_id: str, person_id: str, **kwargs
) -> Optional[JSONDict]:
    """Return a json of controller data."""
    return companies_house_query(
        f"/company/{company_id}/"
        "persons-with-significant-control/"
        f"individual/{person_id}"
    )


def get_significant_controller_company_data(
    company_id: str, firm_id: str, **kwargs
) -> Optional[JSONDict]:
    """Return a json of controller data."""
    return companies_house_query(
        f"/company/{company_id}/"
        "persons-with-significant-control/"
        f"corporate-entity/{firm_id}"
    )


def is_inactive(
    person_data: dict,
    keyword: str = COMPANIES_HOUSE_RESIGNATION_KEYWORD,
    date_format: str = COMPANIES_HOUSE_DATE_FORMAT,
) -> bool:
    """Tests whether board members (default) or controllers are inactive."""
    return (
        keyword in person_data
        and datetime.strptime(person_data[keyword], date_format)
        < datetime.today()
    )


def is_individual_controller_url(url: str) -> bool:
    """Checks if the url implies the controller is a person."""
    return "persons" in url and "individual" in url


def is_person(
    name: str, company_suffixes: Sequence[str] = COMPANY_SUFFIXES
) -> bool:
    """Estimate if board member or controller is a person (company if not)."""
    for suffix in company_suffixes:
        if suffix in name:
            return False
    return True


def filter_active_board_members(g: Graph) -> Graph:
    """Return a graph with only active board members."""
    subgraph = g.copy()
    inactive_board_members = [
        id
        for id, data in subgraph.nodes(data=True)
        if data["bipartite"] == 1 and is_inactive(data["data"])
    ]
    subgraph.remove_nodes_from(inactive_board_members)
    return subgraph


def get_company_category(
    company_id: CompanyIDType,
    uri_codes: Dict[str, str] = COMPANIES_HOUSE_URI_CODES,
) -> str:
    """Return a company category from the first 2 characters."""
    company_id = stringify_company_id(company_id)
    if company_id.isdigit():
        return uri_codes[""]
    try:
        return uri_codes[company_id[:2]]
    except KeyError:
        raise KeyError(f"Company ID Number {company_id} prefix is unlisted.")



"""Basic configuration for adding fixtures for tests.

This adds pytest marks regarding API access to Companies House.
"""

from os import getenv
from typing import Optional

import pytest
from dotenv import load_dotenv
from networkx import Graph

from ukboards.utils import (
    DEFAULT_API_KEY_PATH,
    InternetConnectionError,
    get_external_ip_address,
)

load_dotenv(dotenv_path=DEFAULT_API_KEY_PATH)

COMPANIES_HOUSE_ALLOWED_IP_ADDRESS_NAME: str = (
    "COMPANIES_HOUSE_ALLOWED_IP_ADDRESS"
)
COMPANIES_HOUSE_ALLOWED_IP_ADDRESS: Optional[str] = getenv(
    COMPANIES_HOUSE_ALLOWED_IP_ADDRESS_NAME, ""
)

try:
    IP_ADDRESS: Optional[str] = get_external_ip_address()
except InternetConnectionError:
    IP_ADDRESS = None
    pass

skip_if_not_allowed_ip = pytest.mark.skipif(
    IP_ADDRESS != COMPANIES_HOUSE_ALLOWED_IP_ADDRESS,
    reason=(
        "Fails unless ip address is registered for Companies House api key."
    ),
)


@pytest.fixture
def simple_graph():
    """Return a basic graph for testing input/output."""
    g = Graph()
    g.add_node("034", bipartite=0, name="comp", data={"some": "data"})
    g.add_node("1a", bipartite=1, name="jule", data={"mode": "data-y"})
    g.add_edge("034", "1a", weight=2, data={"edgey": "data"})
    return g



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Utils for saving and loading network files, logging and ordinance data."""

from collections.abc import MutableSequence
from csv import DictReader
from dataclasses import dataclass, field
from datetime import datetime
from json import dump, load
from logging import INFO, FileHandler, getLogger
from os import PathLike
from os.path import getctime
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Final,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import requests
from networkx import (
    Graph,
    node_link_data,
    node_link_graph,
    set_node_attributes,
)
from requests.exceptions import ConnectionError

CHECK_EXTERNAL_IP_ADDRESS_GOOGLE: Final[
    str
] = "https://domains.google.com/checkip"

POSTCODE_IO: Final[str] = "https://api.postcodes.io/"
POSTCODE_CURRENT: Final[str] = POSTCODE_IO + "postcodes/"
POSTCODE_TERMINATED: Final[str] = POSTCODE_IO + "terminated_postcodes/"

DEFAULT_API_KEY_PATH: PathLike = Path(".env")
JSON_DATA_PATH: PathLike = Path("data/json")

LOG_FOLDER: PathLike = Path("logs/")
LOG_TIME_FORMAT: Final[str] = "%a, %d %b %Y %H:%M:%S"
LOG_FILENAME_DATE_FORMAT: Final[str] = "%Y-%m-%d-%H:%M:%S"

JSON_DATE_FORMAT: Final[str] = "%Y-%m-%d %H:%M:%S.%f"
JSON_ADDITIONAL_DATA_KEY: Final[str] = "ukboards-metadata"
METADATA_DATETIME_KEYS: Final[List[str]] = ["start_time", "end_time"]

ORGANISATION_AUXILARY_NODE_DATA_KEY: Final[str] = "org_auxiliary_data"
ORGANISATION_AUXILARY_ATTR_NAMING_DICT_KEY: Final[
    str
] = "data_to_root_atr_mapping"

logger = getLogger(__name__)

CharityAPIKeyType = str
CompanyAPIKeyType = str
APIKeyDictType = Dict[
    str, Optional[Union[CharityAPIKeyType, CompanyAPIKeyType]]
]

QueryValueTypes = Optional[Union[str, bool, int, List]]
QueryParameters = Dict[str, QueryValueTypes]

DataRowDict = Dict[str, Union[str, int]]

JSONDict = Dict[str, Any]

KindsIDType = Set[Optional[Union[str, int]]]
KindsDict = Dict[str, KindsIDType]

CSVRowType = Tuple[int, Dict[str, str]]

NodeKeyType = Union[int, str, float]


def formatted_now_str(date_format: str = LOG_FILENAME_DATE_FORMAT) -> str:
    """Return current time in ``date_format`` format."""
    return datetime.now().strftime(date_format)


DEFAULT_LOG_FILE_NAME = f"default_{formatted_now_str()}.log"


@dataclass
class RunConfig:
    """Base Typed Dict of query runs.

    Todo:
        * Add logs at an INFO default level
        * Consider storing relative log file with time stamp
    """

    # root_id: Optional[Union[str, int]]
    # kind: Literal['charity', 'company']
    parameter_state: QueryParameters = field(default_factory=dict)
    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None
    # logs: [],
    connected_components_count: Optional[int] = None
    kinds_ids_dict: Optional[KindsDict] = None
    # composed_runs: List = field(default_factory=list)


def get_network_json_file_name(
    prefix: str = "charity_",
    time: Optional[datetime] = None,
    time_format: str = LOG_FILENAME_DATE_FORMAT,
) -> str:
    """Return a json file name datetime.now() timestamp by default."""
    time = time or datetime.now()
    return f"{prefix}-{time.strftime(time_format)}.json"


def get_latest_json_file_name(
    prefix: str = "", path: PathLike = JSON_DATA_PATH
) -> PathLike:
    """Return the latest json file in path with prefix."""
    path = Path(path)
    try:
        return max(path.glob(f"*{prefix}*.json"), key=getctime)
    except ValueError:
        raise NoMatchingDataPathError(path=path, prefix=prefix)


def read_csv(path: PathLike, **kwargs) -> Iterator[CSVRowType]:
    """Open `path` and return a list of dicts."""
    with open(path, **kwargs) as csv_file:
        for i, line in enumerate(DictReader(csv_file, **kwargs)):
            yield i, line


def write_json_graph(
    graph: Graph,
    path: PathLike = JSON_DATA_PATH,
    additional_data: JSONDict = None,
    additional_data_key: str = JSON_ADDITIONAL_DATA_KEY,
) -> None:
    """Write a json file including nodes, edges and attributes."""
    path = Path(path)
    path.parent.mkdir(exist_ok=True, parents=True)
    with open(path, "w") as graph_file:
        json_graph: JSONDict = node_link_data(graph)
        if additional_data:
            json_graph[additional_data_key] = additional_data
        dump(json_graph, graph_file, default=str)


def read_json_graph(
    path: PathLike = JSON_DATA_PATH,
    additional_data: bool = False,
    additional_data_key: str = JSON_ADDITIONAL_DATA_KEY,
) -> Union[Graph, Tuple[Graph, Graph]]:
    """Read a json link_data_format file with nodes, edges and attributes."""
    path = Path(path)
    with open(path) as graph_file:
        if additional_data:
            json_graph: JSONDict = load(
                graph_file, object_hook=json_deserialise
            )
            return (
                node_link_graph(json_graph),
                json_graph.get(additional_data_key),
            )
        return node_link_graph(load(graph_file))


def json_deserialise(
    json_dict: JSONDict,
    time_keys: Sequence[str] = METADATA_DATETIME_KEYS,
    time_format: str = JSON_DATE_FORMAT,
    root_key: str = JSON_ADDITIONAL_DATA_KEY,
):
    """Deserialise a JSON Datetime string."""
    for k, v in json_dict.items():
        if k in time_keys:
            json_dict[k] = datetime.strptime(v, time_format)
        elif type(v) is str:
            if v.isdigit() and not v.startswith("0"):
                json_dict[k] = int(v)
            elif v.startswith("{") and v.endswith("}"):
                values = v.strip("{}'").split(", ")
                json_dict[k] = {
                    int(val)
                    if val.isdigit() and not val.startswith("0")
                    else val.strip("'")
                    for val in values
                }
            elif v == "set()":
                json_dict[k] = set()
    return json_dict


def get_ordinance_data(post_code: str) -> Optional[JSONDict]:
    """Request ordinance survey data from postcodes.io."""
    try:
        response: requests.Response = requests.get(
            POSTCODE_CURRENT + post_code
        )
        if response.status_code == 200:
            return response.json()["result"]
        elif response.status_code == 404:
            logger.warning(
                f"ordinance.io query for {post_code} "
                f"returned a 404. Trying {POSTCODE_TERMINATED}"
            )
            response = requests.get(POSTCODE_TERMINATED + post_code)
            if response.status_code == 200:
                return response.json()["result"]
            else:
                logger.error(
                    f"No current or terminated record of {post_code} "
                    f"available at the ordinance survey."
                )
                return None
        else:
            raise
    except ConnectionError:
        raise InternetConnectionError


def ordinance_wrapper(node_key: NodeKeyType, data: JSONDict) -> None:
    """Call ``get_ordinance_data`` on data and return a tuple with node_key."""
    address: Dict[str, str]
    post_code: Optional[str] = None
    if data["kind"] == "company":
        address = data["data"]["company"]["registered_office_address"]
        post_code = address.get("postal_code", None)
    elif data["kind"] == "personal-appointment":
        address = data["data"]["items"]["address"]
        post_code = address.get("postal_code", None)
    elif data["kind"] == "officer":
        address = data["data"]["items"][0]["address"] if data["data"] else None
        post_code = address.get("postal_code", None) if address else None
    elif data["kind"] in ("charity", "trustee"):
        if "Address" in data["data"]:
            address = {
                k: v.strip() if v else v
                for k, v in data["data"]["Address"].items()
            }
            post_code = address.get("Postcode")
            if isinstance(post_code, str):
                post_code = post_code.strip()
    else:
        raise NotImplementedError(f"No implementation of kind: {data['kind']}")
    data["address"] = address
    data["post_code"] = post_code
    data["ordinance"] = get_ordinance_data(post_code) if post_code else None
    data["latitude"] = (
        data["ordinance"]["latitude"] if data["ordinance"] else None
    )
    data["longitude"] = (
        data["ordinance"]["longitude"] if data["ordinance"] else None
    )


def call_node_func(graph: Graph, func: Callable) -> None:
    """Call a function iterating on all graph nodes."""
    try:
        for node in graph.nodes(data=True):
            func(*node)
    except AttributeError:
        logger.warning(
            f"{graph} passed which must be a Graph "
            f"object so {func} cannot be run."
        )


def call_org_sequence_func(
    orgs: MutableSequence,
    graph: Graph,
    func: Callable,
    attr_names_dict: Optional[Dict[str, str]],
    initiate_attr_names: bool = True,
    *args,
    **kwargs,
) -> None:
    """Apply func to orgs with optional graph and return a graph."""
    if initiate_attr_names and attr_names_dict:
        for attr_name in attr_names_dict.values():
            set_node_attributes(graph, None, attr_name)
    for org in orgs:
        func(org, graph, attr_names_dict, *args, **kwargs)


def add_org_auxiliary_data(
    organisation: Any,
    graph: Graph,
    attr_names_dict: Optional[Dict[str, str]],
    data_dict_key: str = ORGANISATION_AUXILARY_NODE_DATA_KEY,
    attr_names_dict_key: str = ORGANISATION_AUXILARY_ATTR_NAMING_DICT_KEY,
) -> None:
    """Add auxiliary organisation data to nodes in graph."""
    for org_id in organisation.active_ids:
        if org_id in graph:
            graph.nodes[org_id][data_dict_key] = organisation.data
            if attr_names_dict:
                for data_attr_name, node_attr_name in attr_names_dict.items():
                    graph.nodes[org_id][node_attr_name] = organisation.data[
                        data_attr_name
                    ]
                graph.nodes[org_id][data_dict_key][
                    attr_names_dict_key
                ] = attr_names_dict


def set_node_data_func(graph: Graph, name: str, func: Callable) -> Graph:
    """Call a function to add data to nodes, iterating on all graph nodes."""
    for node in graph.nodes(data=True):
        node[1][name] = func(*node)


def file_log_handler(
    level: int = INFO,
    filename: str = DEFAULT_LOG_FILE_NAME,
    folder: PathLike = LOG_FOLDER,
    reset_log: bool = True,
    *args,
    **kwargs,
) -> FileHandler:
    """Add a file logger.

    Todo:
        * Consider generalising to type of handler.
    """
    log_path = Path(folder) / filename
    log_path.parent.mkdir(exist_ok=True, parents=True)
    if reset_log:
        file_handler = FileHandler(log_path, mode="w")
    else:
        file_handler = FileHandler(log_path)
    file_handler.setLevel(level)
    return file_handler


def get_kinds_ids_dict(graph: Graph, kinds: Tuple[str, ...]) -> KindsDict:
    """Return a dicts of ids of each kind.

    In the case of Companies House companies there are companies, board members
    and controllers. See ukboards.companies.COMPANEY_NETWORK_KINDS
    """
    kind_dict: KindsDict = {k: set() for k in kinds}
    for node_id, data in graph.nodes(data=True):
        kind_dict[data["kind"]].add(node_id)
    return kind_dict


class Error(Exception):
    """Base class for exceptions in this module.

    See: https://docs.python.org/3/tutorial/errors.html
    """

    pass


class NegativeIntBranchException(Error):
    """Error of a branch being a non-integer and/or less than 0.

    Todo:
        * Consider replacing self.message with self.msg.
    """

    def __init__(
        self,
        branches: int,
        message: str = None,
    ) -> None:
        """Initialise exception attributes based on branch numbers."""
        self.branches = branches
        self.message = message or (
            f"{branches} is an invalid number of "
            "network branches. It must be an int "
            "and > 0."
        )

    def __str__(self) -> str:
        """Set self.__str__ to use self.message."""
        return self.message


class ExceededMaxBranchesException(Error):
    """Error of a branch being a non-integer and/or less than 0.

    Todo:
        * Consider replacing self.message with self.msg.
    """

    def __init__(
        self,
        branches: int,
        max_branches: int,
        message: str = None,
    ) -> None:
        """Initialise exception attributes based on branch numbers."""
        self.branches: int = branches
        self.max_branches: int = max_branches
        self.message: str = message or (
            f"Current branches {branches} >= "
            f"maximum {max_branches} branches "
            "allowed."
        )

    def __str__(self) -> str:
        """Set self.__str__ to use self.message."""
        return self.message


class InternetConnectionError(Error):
    """Either no internet connection or a restricted local network."""

    DEFAULT_MESSAGE = (
        "You external IP address cannot be "
        "found. You may have lost internet "
        "connectivity or have a restricted "
        "local connection without access to "
        "the wider internet, including "
        "google.com, Companies House and "
        "Charity Commission APIs."
    )

    def __init__(self, msg: str = DEFAULT_MESSAGE, *args, **kwargs) -> None:
        """Return normal Error class with custom default msg."""
        super().__init__(msg)


class NoLoadedNetworkDataError(Error):
    """Exception if calculating ego networks prior to loading network data."""

    DEFAULT_MESSAGE = (
        "Network data must be loaded prior to calculating ego networks."
    )

    def __init__(self, msg: str = DEFAULT_MESSAGE, *args, **kwargs) -> None:
        """Add msg = DEFAULT_MESSAGE to standard Exception __init__."""
        super().__init__(msg)


class NoMatchingDataPathError(Error):
    """Exception of no existing path for saving or loading data."""

    def __init__(
        self,
        msg: Optional[str] = None,
        path: PathLike = Path(),
        prefix: str = "",
    ) -> None:
        """Raise error of ``path`` and ``prefix`` failing to match files."""
        self.msg = (
            msg
            or f"No path '{path}' contains files matching prefix: '{prefix}'."
        )
        self.path = path
        self.prefix = prefix
        super().__init__(self.msg)


def get_external_ip_address(
    checkip_url: str = CHECK_EXTERNAL_IP_ADDRESS_GOOGLE,
) -> str:
    """Check external ip address, raise an error if no connection."""
    try:
        return requests.get(checkip_url).text
    except ConnectionError:
        raise InternetConnectionError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test utils."""

from logging import DEBUG, INFO, getLogger
from os import PathLike
from typing import Dict

import pytest
from networkx import Graph

from ukboards.utils import (
    DEFAULT_LOG_FILE_NAME,
    POSTCODE_CURRENT,
    POSTCODE_TERMINATED,
    JSONDict,
    file_log_handler,
    formatted_now_str,
    get_ordinance_data,
    ordinance_wrapper,
    read_csv,
    read_json_graph,
    set_node_data_func,
    write_json_graph,
)

logger = getLogger(__name__)


def test_csv_generator():
    """Test yield of csv rows."""
    orgs = [line for _, line in read_csv("tests/organisation_sample.csv")]
    assert orgs[4]["Company Number"] == "7007198"
    assert orgs[4]["Charity Number"] == "1136495"
    assert orgs[4]["Organisation name"] == "A Space Arts"


def test_read_write_graph(tmp_path, simple_graph):
    """Test writing and reading a graph."""
    test_path = tmp_path / "test_path" / "test_graph.json"
    write_json_graph(simple_graph, test_path)
    f = read_json_graph(test_path)
    assert simple_graph.nodes(data=True) == f.nodes(data=True)
    assert simple_graph.edges == f.edges
    assert list(simple_graph.edges(data=True)) == list(f.edges(data=True))


def test_read_write_metadata(tmp_path, simple_graph):
    """Test reading an example charity graph dataset."""
    TEST_KEY: str = "test-key"
    TEST_ADDITIONAL_DATA: JSONDict = {"some": 5, "more": "data"}
    test_path: PathLike = tmp_path / "test_path" / "test_graph.json"
    write_json_graph(
        simple_graph,
        test_path,
        additional_data=TEST_ADDITIONAL_DATA,
        additional_data_key=TEST_KEY,
    )
    f, d = read_json_graph(
        test_path, additional_data=True, additional_data_key=TEST_KEY
    )
    assert simple_graph.nodes(data=True) == f.nodes(data=True)
    assert simple_graph.edges == f.edges
    assert list(simple_graph.edges(data=True)) == list(f.edges(data=True))
    assert d["some"] == 5


def test_add_file_logger(tmp_path, caplog):
    """Test adding a file logger."""
    test_logger = getLogger("test_logger")
    test_logger.setLevel(DEBUG)

    INFO_LOG_TEXT = "An info level log."
    DEBUG_LOG_TEXT = "A debug level log."
    TEST_FILENAME = "test_log_name.log"
    path_test = tmp_path / TEST_FILENAME

    file_handler = file_log_handler(
        level=INFO,
        filename=TEST_FILENAME,
        folder=tmp_path,
    )
    test_logger.addHandler(file_handler)
    test_logger.info(INFO_LOG_TEXT)
    test_logger.debug(DEBUG_LOG_TEXT)
    log_text = path_test.read_text()
    assert INFO_LOG_TEXT + "\n" == log_text
    assert DEBUG_LOG_TEXT not in log_text
    file_handler2 = file_log_handler(
        level=DEBUG,
        filename=TEST_FILENAME,
        folder=tmp_path,
    )
    test_logger.addHandler(file_handler2)
    test_logger.info(INFO_LOG_TEXT)
    test_logger.debug(DEBUG_LOG_TEXT)
    log_text = path_test.read_text()
    assert log_text == f"{INFO_LOG_TEXT}\n{DEBUG_LOG_TEXT}\n"
    file_handler3 = file_log_handler(
        level=DEBUG,
        filename=TEST_FILENAME,
        folder=tmp_path,
        reset_log=True,
    )
    test_logger.addHandler(file_handler3)
    log_text = path_test.read_text()
    assert log_text == ""


def test_query_address(requests_mock, caplog):
    """Test querying ordinance data from punchdrunk postcode."""
    LAT: float = 51.590792
    LON: float = -0.06056
    POSTCODE: str = "N17 9LH"
    URL_CONVERTED_POSTCODE: str = POSTCODE.replace(" ", "%20")
    response: dict = {
        "status": 200,
        "result": {
            "postcode": POSTCODE,
            "quality": 1,
            "eastings": 534449,
            "northings": 189775,
            "country": "England",
            "nhs_ha": "London",
            "longitude": LON,
            "latitude": LAT,
            "european_electoral_region": "London",
        },
    }
    requests_mock.get(POSTCODE_CURRENT + URL_CONVERTED_POSTCODE, json=response)
    ordinance_data = get_ordinance_data(POSTCODE)
    assert ordinance_data["postcode"] == POSTCODE
    assert ordinance_data["latitude"] == LAT
    assert ordinance_data["longitude"] == LON


def test_query_address_terminated(caplog):
    """Test querying ordinance data from punchdrunk postcode."""
    TERMINATED_POSTCODE: str = "WC1R 4GB"
    YEAR_TERMINATED: int = 2016
    MONTH_TERMINATED: int = 3
    LAT: float = 51.518359
    LON: float = -0.117027
    ordinance_data = get_ordinance_data(TERMINATED_POSTCODE)
    assert ordinance_data["postcode"] == TERMINATED_POSTCODE
    assert ordinance_data["latitude"] == LAT
    assert ordinance_data["longitude"] == LON
    assert ordinance_data["year_terminated"] == YEAR_TERMINATED
    assert ordinance_data["month_terminated"] == MONTH_TERMINATED
    assert caplog.messages == [
        f"ordinance.io query for {TERMINATED_POSTCODE} returned a "
        f"404. Trying {POSTCODE_TERMINATED}"
    ]


def test_incorrect_post_code(caplog):
    """Test querying ordinance data from incorrect postcode."""
    INCORRECT_POSTCODE: str = "WC2 9PA"
    ordinance_data = get_ordinance_data(INCORRECT_POSTCODE)
    assert ordinance_data is None
    assert caplog.messages == [
        (
            f"ordinance.io query for {INCORRECT_POSTCODE} returned a "
            f"404. Trying {POSTCODE_TERMINATED}"
        ),
        (
            f"No current or terminated record of {INCORRECT_POSTCODE} "
            f"available at the ordinance survey."
        ),
    ]


def test_ordinance_wrapper_no_additional_company_data(caplog):
    """Test edge company case where no additional data is available.

    Todo:
        * {'address_line_1': '35-47 Bethnal Green Road', 'locality': 'London',
           'country': 'United Kingdom', 'premises': 'Unit 2.E.03'}
    """
    node_tuple: tuple[str, dict] = (
        "G8Y9TpkksSLxGzzWMeixqNO15ow",
        {"kind": "officer", "bipartite": 1, "data": None},
    )
    ordinance_wrapper(*node_tuple)
    assert node_tuple[1]["address"] is None
    assert node_tuple[1]["post_code"] is None
    assert node_tuple[1]["ordinance"] is None
    assert node_tuple[1]["latitude"] is None
    assert node_tuple[1]["longitude"] is None


@pytest.mark.remote_data
def test_ordinance_wrapper_no_null_charity_trustee_address(caplog):
    """Test edge charity case where no address data is available."""
    NULL_ADDRESS_DICT = {
        "Line1": None,
        "Line2": None,
        "Line3": None,
        "Line4": None,
        "Line5": None,
        "Postcode": None,
    }
    node_tuple: tuple[str, dict] = (
        "1075794",
        {
            "kind": "trustee",
            "bipartite": 1,
            "data": {"Address": NULL_ADDRESS_DICT},
        },
    )
    ordinance_wrapper(*node_tuple)
    assert node_tuple[1]["address"] == NULL_ADDRESS_DICT
    assert node_tuple[1]["post_code"] is None
    assert node_tuple[1]["ordinance"] is None
    assert node_tuple[1]["latitude"] is None
    assert node_tuple[1]["longitude"] is None


def test_set_node_data_func():
    """Test setting a new component of data in a network."""
    g: Graph = Graph()
    ATTR_NAME: str = "9times"
    g.add_nodes_from([1, 2])

    def n_times_9(n: int, data: Dict) -> int:
        return n * 9

    set_node_data_func(g, ATTR_NAME, n_times_9)
    assert list(g.nodes(data=True)) == [
        (1, {ATTR_NAME: 9}),
        (2, {ATTR_NAME: 18}),
    ]


def test_formatted_now_str():
    """Test returning formatted str of current dateime."""
    name_1 = f"default_{formatted_now_str()}.log"  # Later because of import
    name_2 = DEFAULT_LOG_FILE_NAME  # Early because of the import
    assert name_1[:17] == name_2[:17]
    assert name_1[27:] == name_2[27:] == ".log"
    assert name_1[17:27] >= name_2[17:27]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Base mode for querying Charity Commission SOAP API."""

from dataclasses import dataclass, field
from logging import getLogger
from os import getenv
from typing import Final, List, Optional, Tuple, Union

from dotenv import load_dotenv
from networkx import Graph, compose, is_bipartite
from zeep import Client, Plugin, Settings
from zeep.exceptions import Fault
from zeep.helpers import serialize_object

from .utils import DEFAULT_API_KEY_PATH, CharityAPIKeyType, RunConfig

logger = getLogger(__name__)

load_dotenv(dotenv_path=DEFAULT_API_KEY_PATH)

CHARITY_COMMISSION_WSDL = (
    "https://apps.charitycommission.gov.uk/"
    "Showcharity/API/SearchCharitiesV1/"
    "SearchCharitiesV1.asmx?wsdl"
)

CHARITY_COMMISSION_API_KEY_NAME: Final[str] = "APIKey"
CHARITY_COMMISSION_API_KEY_ENV_NAME: Final[str] = "CHARITY_COMMISSION_KEY"
CHARITY_COMMISSION_API_KEY: Final[Optional[CharityAPIKeyType]] = getenv(
    CHARITY_COMMISSION_API_KEY_ENV_NAME
)

CHARITY_NETWORK_KINDS: Final[Tuple[str, str]] = ("charity", "trustee")

CharityIDType = int
CharityRootIDsType = Union[CharityIDType, Tuple[CharityIDType, ...]]


class CharitiesAuthPlugin(Plugin):
    """Add an APIKey to each request."""

    def __init__(
        self,
        api_key_name: str = CHARITY_COMMISSION_API_KEY_NAME,
        api_key_value: Optional[
            CharityAPIKeyType
        ] = CHARITY_COMMISSION_API_KEY,
    ) -> None:
        """Initialise api_key_name and api_key_value."""
        self.api_key_name = api_key_name
        self.api_key_value = api_key_value

    def egress(self, envelope, http_headers, operation, binding_options):
        """Auto add to the envelope (or replace) Charity api_key.

        :param envelope: The envelope as XML node
        :param http_headers: Dict with the HTTP headers
        :param operation: The associated Operation instance
        :param binding_options: Binding specific options for the operation
        """
        for element in operation.input.body.type.elements:
            if (
                element[0] == self.api_key_name
                and element[1].name == self.api_key_name
            ):
                key_type = element[1]
                key_type.render(envelope[0][0], self.api_key_value)
        return envelope, http_headers


def get_client(
    wsdl: str = CHARITY_COMMISSION_WSDL,
    raw_response: bool = False,
    api_key_name: str = CHARITY_COMMISSION_API_KEY_NAME,
    api_key_value: Optional[CharityAPIKeyType] = CHARITY_COMMISSION_API_KEY,
    plugins: List[Plugin] = None,
    **kwargs,
) -> Client:
    """Generate a Client for querying Charities Commision API."""
    settings: Settings = Settings(
        strict=False, xml_huge_tree=True, raw_response=raw_response
    )
    plugins = (
        [CharitiesAuthPlugin(api_key_name, api_key_value)]
        if not plugins
        else plugins
    )
    return Client(wsdl=wsdl, settings=settings, plugins=plugins, **kwargs)


def check_registered_charity_number(
    name: str,
    charity_number: CharityIDType,
    client: Optional[Client] = None,
) -> int:
    """Check if registered charity_number matches searched name."""
    if not client:
        client = get_client()
    charities = client.service.GetCharitiesByName(
        strSearch=name,
    )
    for charity in charities:
        charity_data = client.service.GetCharityByRegisteredCharityNumber(
            registeredCharityNumber=charity["RegisteredCharityNumber"],
        )
        if charity_data["CharityNumber"] == charity_number:
            return charity_data["RegisteredCharityNumber"]
    raise Exception(
        'No charity found with "CharityNumber": ' f"{charity_number}"
    )


def get_charity_network(
    charity_number: CharityIDType = 1085314,  # TATE
    branches: int = 0,  # FOUNDATION
    client: Optional[Client] = None,
    api_key: Optional[str] = CHARITY_COMMISSION_API_KEY,
    test_name: str = None,
    *args,
    **kwargs,
) -> Optional[Graph]:
    """Query Charity Commission API for board interlock network."""
    g: Graph = Graph()
    if not client:
        client = get_client(api_key_value=api_key)
    assert isinstance(charity_number, int)
    try:
        charity_data = client.service.GetCharityByRegisteredCharityNumber(
            registeredCharityNumber=charity_number,
        )
    except Fault:
        logger.error(f"Fault error pulling for {charity_number}")
        return None
    if not charity_data:
        logger.warning(f"No data on charity {charity_number}")
        if not test_name:
            logger.warning(
                "No records from Charities Commision on "
                "this Arts Council Instition"
            )
            return None
    charity_name = charity_data["CharityName"]
    if test_name:  # Test the name of the queried Charity is the intended
        try:
            assert charity_name == test_name
        except AssertionError:
            logger.exception(
                'Referral test name "{0}" is different from '
                '"{1}" which is associated '
                "with charity_number {2}".format(
                    test_name.strip(), charity_name.strip(), charity_number
                )
            )
    g.add_node(
        charity_number,
        name=charity_name.strip(),
        bipartite=0,
        # Add a charity/company attribute
        kind=CHARITY_NETWORK_KINDS[0],
        data=serialize_object(charity_data),
    )
    logger.debug(charity_name)
    for subsidiary in range(charity_data["SubsidiaryNumber"] + 1):
        trustees = client.service.GetCharityTrustees(
            registeredCharityNumber=charity_number, subsidiaryNumber=subsidiary
        )
        if not trustees:
            logger.warning(
                "No trustees for charity {0} ({1} "
                "subsidiary {2})".format(
                    charity_name.strip(), charity_number, subsidiary
                )
            )
            continue
        for trustee in trustees:
            logger.debug(
                "{0} {1} {2}".format(
                    charity_number,
                    trustee["TrusteeName"],
                    trustee["TrusteeNumber"],
                )
            )
            g.add_node(
                trustee["TrusteeNumber"],
                name=trustee["TrusteeName"].strip(),
                bipartite=1,
                kind=CHARITY_NETWORK_KINDS[1],
                data=serialize_object(trustee),
            )
            g.add_edge(charity_number, trustee["TrusteeNumber"])
            if branches and trustee["RelatedCharitiesCount"]:
                for charity in trustee["RelatedCharities"]:
                    # related_charity_number = \
                    #     get_charity_number(charity['CharityName'],
                    #                                   charity['CharityNumber'],
                    #                                   client=client)
                    if charity["CharityNumber"] not in g.nodes:
                        subgraph = get_charity_network(
                            charity["CharityNumber"],
                            branches - 1,
                            client=client,  # type: ignore[misc]
                            test_name=charity["CharityName"],
                            *args,
                            **kwargs,
                        )
                        assert is_bipartite(subgraph)
                        g = compose(g, subgraph)
                        assert is_bipartite(g)
    return g


@dataclass
class CharityRunConfig(RunConfig):
    """RunConfig dict with type enforced root_id."""

    root_id: Optional[CharityRootIDsType] = None
    success: Optional[bool] = False
    kind: Final[str] = "charity"
    composed_runs: List["CharityRunConfig"] = field(default_factory=list)



# -*- coding: utf-8 -*-

"""Console script for ukboards."""

from json import dumps
from logging import getLogger
from sys import exit
from typing import Optional

import click
from dotenv import dotenv_values
from zeep import Client

from .charities import (
    CHARITY_COMMISSION_API_KEY,
    CHARITY_COMMISSION_API_KEY_ENV_NAME,
    CharityIDType,
    get_client,
)
from .companies import (
    COMPANIES_HOUSE_API_KEY,
    COMPANIES_HOUSE_API_KEY_ENV_NAME,
    CompanyIDType,
    companies_house_query,
)
from .utils import (
    DEFAULT_API_KEY_PATH,
    APIKeyDictType,
    CharityAPIKeyType,
    CompanyAPIKeyType,
    JSONDict,
)

logger = getLogger(__name__)


@click.group(name="ukboards")
@click.option(
    "--api-keys-path",
    "-a",
    "api_keys_path",
    type=click.Path(dir_okay=False, file_okay=True),
    envvar="UK_BOARD_KEYS",
    default=DEFAULT_API_KEY_PATH,
    nargs=1,
    help="Path to file with API keys (default=.env)",
)
@click.option(
    "--indent",
    "-i",
    type=int,
    nargs=1,
    default=2,
    show_default=True,
    help="How many spaces to indent printing json queries.",
)
@click.pass_context
def ukboards(
    ctx: click.Context, indent: int, api_keys_path: click.Path
) -> int:
    """Query UK company and charity board data."""
    ctx.ensure_object(dict)
    if api_keys_path:
        try:
            api_keys: APIKeyDictType = dotenv_values(api_keys_path)
            for key_name, key in api_keys.items():
                ctx.obj[key_name] = key
        except FileNotFoundError:
            pass
            # Todo: * Add opitions for generating an .env file for API keys.
            # path = input("No file provided for Companies House or "
            #              "Charities Commission API keys. Press enter to "
            #              f"create a default '{DEFAULT_API_KEY_PATH}' file "
            #              "in this folder or type your preferred filename "
            #              "(or CTL+C to quit): ")
    ctx.obj["indent"] = indent
    return 0


@ukboards.command()
@click.argument("charity_id", type=CharityIDType, nargs=1)
@click.option(
    "--api-key",
    "-k",
    "api_key",
    type=CharityAPIKeyType,
    nargs=1,
    default=CHARITY_COMMISSION_API_KEY,
)
@click.pass_context
def charity(
    ctx: click.Context, charity_id: CharityIDType, api_key: CharityAPIKeyType
) -> None:
    """Query Charities Commision by charity id."""
    if CHARITY_COMMISSION_API_KEY_ENV_NAME in ctx.obj:
        api_key = ctx.obj[CHARITY_COMMISSION_API_KEY]
    client: Client = get_client(api_key_value=api_key)
    charity_data = client.service.GetCharityByRegisteredCharityNumber(
        registeredCharityNumber=charity_id,
    )
    click.echo(charity_data)


@ukboards.command()
@click.argument("company_id", type=CompanyIDType, nargs=1)
@click.option(
    "--api-key",
    "-k",
    "api_key",
    type=CompanyAPIKeyType,
    nargs=1,
    default=COMPANIES_HOUSE_API_KEY,
)
@click.pass_context
def company(
    ctx: click.Context, company_id: CompanyIDType, api_key: CompanyAPIKeyType
) -> None:
    """Query Companies House by company id."""
    if COMPANIES_HOUSE_API_KEY_ENV_NAME in ctx.obj:
        api_key = ctx.obj[COMPANIES_HOUSE_API_KEY_ENV_NAME]
    company_json: Optional[JSONDict] = companies_house_query(
        f"/company/{company_id}", auth_key=api_key
    )
    click.echo(dumps(company_json, indent=ctx.obj["indent"]))


@ukboards.command()
@click.argument("csv_path", type=click.Path(exists=True), nargs=1)
def csv_organisations(csv_path: click.Path):
    """Path to csv with company and charity ids."""
    pass


if __name__ == "__main__":
    exit(ukboards(obj={}))  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test querying Charities Commision API."""

from typing import Final
from xml.etree import ElementTree

import pytest
from networkx import is_connected
from networkx.algorithms import bipartite
from requests import Response
from zeep import Client
from zeep.plugins import HistoryPlugin

from ukboards.charities import (
    CHARITY_COMMISSION_WSDL,
    CharitiesAuthPlugin,
    CharityIDType,
    check_registered_charity_number,
    get_charity_network,
    get_client,
)
from ukboards.utils import CharityAPIKeyType

CHARITY_COMMISSION_API: Final[str] = CHARITY_COMMISSION_WSDL.split("?")[0]

TEST_PHOTOGRAPHERS_GALLERY_NAME: Final[str] = (
    "THE PHOTOGRAPHERS' GALLERY LTD                    "
    "                                                  "
    "                                                  "
)

TEST_TAG_NS: Final[str] = "http://www.charitycommission.gov.uk/"

TEST_PHOTOGRAPHERS_CHARITY_NUMBER: Final[int] = 262548

TEST_API_KEY: Final[str] = "A-fake-test-key"

TATE_FOUNDATION_ID: Final[CharityIDType] = 1085314
TATE_FOUNDATION_NAME: Final[str] = "TATE FOUNDATION"


@pytest.fixture
def test_client(requests_mock, maxlen: int = 20):
    """A mock test client using the 19/12/2019 Charties Commission API."""
    history: HistoryPlugin = HistoryPlugin(maxlen=maxlen)
    auth: CharityAPIKeyType = CharitiesAuthPlugin(api_key_value=TEST_API_KEY)
    with open("tests/charities_api.wsdl", "r") as charities_api:
        requests_mock.get(CHARITY_COMMISSION_WSDL, text=charities_api.read())
        client: Client = get_client(plugins=[history, auth])
        assert hasattr(client.service, "GetCharitiesByName")
        return client


@pytest.fixture
@pytest.mark.remote_data
def history_client(maxlen: int = 20):
    """A client with a history plugin, requires a real api key in a .env."""
    history: HistoryPlugin = HistoryPlugin(maxlen=maxlen)
    client: Client = get_client(plugins=[history, CharitiesAuthPlugin()])
    assert hasattr(client.service, "GetCharitiesByName")
    return client


class TestZeepCharityClient:
    """Test generating a zeep `Client` object for Charity Commision API."""

    def test_specify_config(self, requests_mock):
        """Basic test of generating correct Client configuration."""
        TEST_WSDL_CODE: str = "http://a-test-api-code.uk/apiTest.asmx?wsdl"

        TEST_SOAP_ENV: str = """\
        <soapenv:Envelope
            xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"
            xmlns:v1="http://schemas.conversesolutions.com/xsd/dmticta/v1">
        <soapenv:Header/>
        <soapenv:Body>
        <v1:GetVehicleLimitedInfo>
        <v1:vehicleNo>?</v1:vehicleNo>
        <v1:phoneNo>?</v1:phoneNo>
        </v1:GetVehicleLimitedInfo>
        </soapenv:Body>
        </soapenv:Envelope>"""

        requests_mock.get(TEST_WSDL_CODE, text=TEST_SOAP_ENV)
        test_client: Client = get_client(TEST_WSDL_CODE)
        assert test_client.wsdl.location == TEST_WSDL_CODE
        assert test_client.settings.strict is False
        assert test_client.settings.xml_huge_tree is True

    def test_mock_get_charity_by_name(self, requests_mock, test_client):
        """Test mock query of GetCharitiesByName."""
        TEST_EMAIL: str = "an.email@test.org.uk"
        TEST_PHONE_NUMBER: str = "02075555555"

        TEST_SOAP_RESPONSE: str = f"""\
        <?xml version="1.0" encoding="utf-8"?>
        <soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/"
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
            xmlns:xsd="http://www.w3.org/2001/XMLSchema">
          <soap:Body>
            <GetCharitiesByNameResponse xmlns="{TEST_TAG_NS}">
              <GetCharitiesByNameResult>
                <CharityList>
                  <RegisteredCharityNumber>
                    {TEST_PHOTOGRAPHERS_CHARITY_NUMBER}
                  </RegisteredCharityNumber>
                  <SubsidiaryNumber>
                    0
                  </SubsidiaryNumber>
                  <CharityName>
                    {TEST_PHOTOGRAPHERS_GALLERY_NAME}
                  </CharityName>
                  <MainCharityName>
                    {TEST_PHOTOGRAPHERS_GALLERY_NAME}
                  </MainCharityName>
                  <RegistrationStatus>
                    Registered
                  </RegistrationStatus>
                  <PublicEmailAddress>
                    {TEST_EMAIL}
                  </PublicEmailAddress>
                  <MainPhoneNumber>
                    {TEST_PHONE_NUMBER}
                  </MainPhoneNumber>
                </CharityList>
              </GetCharitiesByNameResult>
            </GetCharitiesByNameResponse>
          </soap:Body>
        </soap:Envelope>"""

        requests_mock.post(CHARITY_COMMISSION_API, text=TEST_SOAP_RESPONSE)
        response: Response = test_client.service.GetCharitiesByName(
            strSearch=TEST_PHOTOGRAPHERS_GALLERY_NAME
        )
        assert response[0]["CharityName"].strip() == (
            TEST_PHOTOGRAPHERS_GALLERY_NAME.rstrip()
        )
        assert response[0]["RegisteredCharityNumber"] == (
            TEST_PHOTOGRAPHERS_CHARITY_NUMBER
        )
        assert response[0]["PublicEmailAddress"].strip() == TEST_EMAIL
        assert response[0]["MainPhoneNumber"].strip() == TEST_PHONE_NUMBER

        history = test_client.plugins[0]
        assert history.last_received["envelope"][0][0].tag.endswith(
            "GetCharitiesByNameResponse"
        )
        assert len(history._buffer) == 1

    @pytest.mark.remote_data
    def test_get_charity_by_name(self, history_client):
        """Test raw_response and history of GetCharitiesByName query."""
        history = history_client.plugins[0]
        with history_client.settings(raw_response=True):
            raw_response = history_client.service.GetCharitiesByName(
                strSearch="TATE GALLERY FOUNDATION"
            )
            assert raw_response.status_code == 200
            with open("tests/charity_soap.wsdl", "rb") as test_output:
                test_tree = ElementTree.fromstring(test_output.read())
                resp_tree = ElementTree.fromstring(raw_response.content)
                assert test_tree.tag == resp_tree.tag
                for test in test_tree.iter():
                    for resp in resp_tree.iter():
                        if test.tag == resp.tag:
                            break
                    if resp.text:
                        assert resp.text.strip() == test.text.strip()
        assert history.last_sent["envelope"][0][0].tag.endswith(
            "GetCharitiesByName"
        )
        assert len(history._buffer) == 1
        response = history_client.service.GetCharitiesByName(
            strSearch="TATE GALLERY FOUNDATION"
        )
        assert response[0]["CharityName"].rstrip() == (
            "THE TATE GALLERY FOUNDATION"
        )
        assert history.last_received["envelope"][0][0].tag.endswith(
            "GetCharitiesByNameResponse"
        )
        assert len(history._buffer) == 2


class TestGetCharityNumber:
    """Test querying basic charity data."""

    @pytest.mark.remote_data
    def test_check_registered_charity_number(self):
        """Test on Photography's Gallery Limited."""
        charity = check_registered_charity_number(
            TEST_PHOTOGRAPHERS_GALLERY_NAME, TEST_PHOTOGRAPHERS_CHARITY_NUMBER
        )
        # charity = check_registered_charity_number(correct_charity_name,
        #                                           correct_charity_number,
        #                                           client=history_client)
        assert charity == TEST_PHOTOGRAPHERS_CHARITY_NUMBER


class TestGetCharityNetwork:
    """Test Contructing a Charity Network."""

    @pytest.mark.remote_data
    def test_tate_foundation(self, history_client):
        """Test basic query of Tate_Foundation board members."""
        charity_network = get_charity_network(
            TATE_FOUNDATION_ID, client=history_client
        )
        assert len(charity_network) == 15
        assert is_connected(charity_network)
        tate_foundation, board_members = bipartite.sets(charity_network)
        assert len(board_members) == 14
        assert (
            charity_network.nodes[TATE_FOUNDATION_ID]["name"]
            == TATE_FOUNDATION_NAME
        )
        assert charity_network.nodes[11965582]["name"] == (
            "Dr Maria " "Jane Balshaw"
        )

    @pytest.mark.remote_data
    def test_tate_foundation_str_id(self, history_client):
        """Test basic query of Tate_Foundation board members."""
        charity_network = get_charity_network("1085314", client=history_client)
        assert len(charity_network) == 15
        assert is_connected(charity_network)
        tate_foundation, board_members = bipartite.sets(charity_network)
        assert len(board_members) == 14
        assert type(list(tate_foundation)[0]) is int

    @pytest.mark.remote_data
    def test_1hop_tate(self, history_client):
        """Test 1 hop query of Tate_Foundation board members."""
        charity_network = get_charity_network(
            1085314, branches=1, client=history_client
        )
        assert len(charity_network) == 73
        assert is_connected(charity_network)
        tate_foundation, board_members = bipartite.sets(charity_network)
        assert len(board_members) == 64



# -*- coding: utf-8 -*-

"""Main module."""

from collections.abc import MutableSequence
from dataclasses import dataclass, field
from datetime import datetime
from logging import DEBUG, INFO, Logger, StreamHandler, getLogger
from os import PathLike
from pathlib import Path
from sys import stdout
from typing import (
    Callable,
    ClassVar,
    Dict,
    Final,
    Generator,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypedDict,
    Union,
    overload,
)

from networkx import (
    Graph,
    NodeNotFound,
    compose_all,
    ego_graph,
    number_connected_components,
)

from .charities import (
    CHARITY_NETWORK_KINDS,
    CharityIDType,
    CharityRunConfig,
    get_charity_network,
)
from .companies import (
    CompanyIDType,
    CompanyNetworkClient,
    CompanyRunConfig,
    stringify_company_id,
)
from .utils import (
    JSON_DATA_PATH,
    LOG_TIME_FORMAT,
    CSVRowType,
    DataRowDict,
    NoLoadedNetworkDataError,
    QueryParameters,
    file_log_handler,
    get_kinds_ids_dict,
    get_latest_json_file_name,
    get_network_json_file_name,
    read_csv,
    read_json_graph,
    write_json_graph,
)

logger: Logger = getLogger(__name__)
logger.setLevel(DEBUG)
stream_handler: StreamHandler = StreamHandler(stdout)
stream_handler.setLevel(INFO)
logger.addHandler(stream_handler)

DEFAULT_COMPANY_COLUMN_NAME: Final[str] = "Company"
DEFAULT_CHARITY_COLUMN_NAME: Final[str] = "Charity"
DEFAULT_ORGANISATION_COLUMN_NAME: Final[str] = "Organisation"


OrganisationIDsType = Union[CharityIDType, CompanyIDType]
OrganisationsDataReaderType = Callable[..., Iterator[CSVRowType]]
OrganisationsDataReaderParamsType = Dict


@dataclass
class OrganisationEntry:
    """Basic data structure for entities of companies and/or charities."""

    data: DataRowDict
    organisation_key_name: str = DEFAULT_ORGANISATION_COLUMN_NAME
    charity_key_name: str = DEFAULT_CHARITY_COLUMN_NAME
    company_key_name: str = DEFAULT_COMPANY_COLUMN_NAME
    _skip_charity: bool = False
    _skip_company: bool = False

    def __str__(self):
        """Return str of entity's name, company and charity ids."""
        return (
            f"{self.name}: Company {self.company_id} | "
            f"Charity {self.charity_id}"
        )

    @property
    def active_ids(self) -> Generator[OrganisationIDsType, None, None]:
        """Return active IDs of self."""
        if self.charity_id and not self._skip_charity:
            yield self.charity_id
        if self.company_id and not self._skip_company:
            yield self.company_id

    @property
    def company_id(self) -> Optional[CompanyIDType]:
        """Return and cash entity's company_id if it exists."""
        if not hasattr(self, "_company_id"):
            self._company_id: Optional[CompanyIDType]
            if self.company_key_name in self.data:
                self._company_id = stringify_company_id(
                    self.data[self.company_key_name]
                )
            else:
                return None
        return self._company_id

    @property
    def charity_id(self) -> Optional[CharityIDType]:
        """Return and cash entity's charity_id if it exists.

        Todo:
            * Refactor for efficiency
        """
        if not hasattr(self, "_charity_id"):
            self._charity_id: Optional[CharityIDType]
            if self.charity_key_name in self.data:
                raw_id: Union[int, str] = self.data[self.charity_key_name]
                if raw_id:
                    if isinstance(raw_id, int):
                        # Todo: assess if this will only return ''
                        self._charity_id = raw_id
                    elif isinstance(raw_id, str) and raw_id.isdigit():
                        self._charity_id = int(raw_id)
                    else:
                        logger.debug(
                            f"{raw_id} is not an int. "
                            f"Skipping for {self} charity_id."
                        )
                else:
                    self._charity_id = None
            else:
                self._charity_id = None
        return self._charity_id

    @property
    def name(self):
        """Return and cash entity's name."""
        if not hasattr(self, "_name"):
            if self.organisation_key_name in self.data:
                self._name = self.data[self.organisation_key_name]
        return self._name


OrganisationSequenceList = List[OrganisationEntry]

# Todo: Replace OrganisationNetworkSequence hinting with TypedDict
# OrganisationNetworkSequence = Sequence[
#     Dict[str, Optional[Union[Graph, OrganisationEntry, OrganisationIDsType]]]
# ]


class OrganisationNetworkDict(TypedDict):
    """Network generated and data from ``organisation`` as seed node."""

    network: Graph
    organisation: OrganisationEntry
    organisation_id: OrganisationIDsType
    organisation_type: str


@dataclass
class OrganisationSequence(MutableSequence):
    """Class for managing queires of a Sequence of ``OrganisationEntry``.

    Todo:
        * Add options for managing additional **kwargs:
          https://stackoverflow.com/questions/54927763/
          control-initialize-order-when-python-dataclass-inheriting-a-class
        * Refactor charities query to generate a client like
          ``CompanyNetworkClient``
        * Include reset options in various cachable attributes
        * Fix ambiguity of self._company_runs vs self._company_composed_runs
    """

    organisations_data_source: Optional[PathLike] = None
    organisations_entry_params: DataRowDict = field(default_factory=dict)
    company_client_params: QueryParameters = field(default_factory=dict)
    charity_client_params: QueryParameters = field(default_factory=dict)
    log_params: QueryParameters = field(default_factory=dict)
    reset_logs: bool = True

    # _CONFIG_ATTRS: Final[Tuple[QueryParameters, ...]] = (data_reader_params,
    #         organisation_entry_params, charity_client_params,
    #         company_client_params, log_params)

    _data_reader: OrganisationsDataReaderType = read_csv
    _data_reader_params: OrganisationsDataReaderParamsType = field(
        default_factory=dict
    )
    _charity_runs: List[CharityRunConfig] = field(default_factory=list)
    _company_runs: List[CompanyRunConfig] = field(default_factory=list)
    _charity_composed_runs: List[Graph] = field(default_factory=list)
    _company_composed_runs: List[Graph] = field(default_factory=list)
    _charity_networks_cached: bool = False
    _company_networks_cached: bool = False
    __company_client_class: ClassVar[
        Type[CompanyNetworkClient]
    ] = CompanyNetworkClient
    # __charity_client_class  # To be added when charity query is refactored
    __charity_client: ClassVar[Callable] = get_charity_network

    # def __aiter__(self):
    #     return self

    # async def __anext__(self):
    #     raise StopAsyncIteration

    # Consider internalising active_charity_ids and active_company_ids

    def __str__(self) -> str:
        """Return summary of included organisations and state of queries."""
        active_charities: int = len(list(self.active_charity_ids))
        total_charities: int = len(list(self.charity_ids))
        active_companies: int = len(list(self.active_company_ids))
        total_companies: int = len(list(self.company_ids))
        return (
            f"{len(self)} UK Organisations: "
            f"{active_charities}/{total_charities} Charities active | "
            f"{active_companies}/{total_companies} Companies active"
        )

    @overload
    def __getitem__(self, index: int) -> OrganisationEntry:
        """Return entity at index position."""
        ...

    @overload
    def __getitem__(self, index: slice) -> List[OrganisationEntry]:
        """Return range of entities across slice of index positions."""
        ...

    def __getitem__(
        self, index: Union[int, slice]
    ) -> Union[OrganisationEntry, List[OrganisationEntry]]:
        """Return entity at index positions in self.organisations list."""
        return self.organisations[index]

    def __setitem__(self, index, value) -> None:
        """Set entity at index position in self.organisations list."""
        self.organisations[index] = value

    def __delitem__(self, index) -> None:
        """Delete entity at index position in self.organisations list."""
        del self.organisations[index]

    def __len__(self) -> int:
        """Return length of self.organisations list."""
        return len(self.organisations)

    # def __new__(cls, *args, **kwargs):
    #     obj = object.__new__(cls)
    #     for attr in cls._CONFIG_ATTRS:
    #         setattr(obj, attr, {**attr, **kwargs})
    #     super().__init__(obj, *args, **kwargs)
    #     assert False
    #     return obj

    # Todo: consider auto loading organisation data here
    # def __post_init__(self, *args, **kwargs) -> None:
    #     """Unpack kwargs and pass to relevant elements."""
    #     assert False
    #     for attr in self._CONFIG_ATTRS:
    #         setattr(self, attr, **kwargs)

    def insert(self, index, value) -> None:
        """Insert value to index position of self.organisations list."""
        self.organisations.insert(index, value)

    @property
    def organisations(
        self,  # *args: QueryParameters, **kwargs: QueryParameters
    ) -> OrganisationSequenceList:
        """Return list of self._organisations and cache from params.

        Todo:
            * Refactor for adding other parameters via a separate method
            * Refactor ``data_reader`` to be managed separately
        """
        if not hasattr(self, "_organisations"):
            # reader_dict_args: QueryParameters = {
            #     **self.data_reader_params,
            #     # **kwargs,
            # }
            # org_entry_args: QueryParameters = {
            #     **self.organisation_entry_params,
            #     # **kwargs,
            # }
            # self._organisations: OrganisationSequenceList = [
            #     OrganisationEntry(org, **org_entry_args)
            #     # for _, org in self.data_reader(*args, **reader_dict_args)
            #     for _, org in self.data_reader(**reader_dict_args)
            self._organisations: OrganisationSequenceList = [
                OrganisationEntry(
                    {**org, "line_number": line_number},
                    # https://github.com/python/mypy/issues/5382[arg-type]
                    **self.organisations_entry_params,  # type: ignore
                )
                # for _, org in self.data_reader(*args, **reader_dict_args)
                for line_number, org in self._data_reader(
                    self.organisations_data_source, **self._data_reader_params
                )
            ]
        return self._organisations

    @property
    def active_charity_ids(self) -> Generator[CharityIDType, None, None]:
        """Return charity ids that contingent on _skip_charity."""
        for organisation in self:
            if organisation.charity_id and not organisation._skip_charity:
                yield organisation.charity_id

    @property
    def active_company_ids(self) -> Generator[CompanyIDType, None, None]:
        """Return company ids that contingent on _skip_company."""
        for organisation in self:
            if organisation.company_id and not organisation._skip_company:
                yield organisation.company_id

    @property
    def charity_ids(self) -> Generator[Optional[CharityIDType], None, None]:
        """Return all charity ids listed in organisations.

        Todo:
            * Consider refactoring iteraction to take advantage.
        """
        for organisation in self:
            yield organisation.charity_id

    @property
    def company_ids(self) -> Generator[Optional[CompanyIDType], None, None]:
        """Return all company ids listed in organisations.

        Todo:
            * Consider refactoring iteraction to take advantage.
        """
        for organisation in self:
            yield organisation.company_id

    @property
    def company_client(
        self,  # *args: QueryParameters, **kwargs: QueryParameters
    ) -> Graph:
        """Cache and return results from _company_client method."""
        if not hasattr(self, "_company_client"):
            # fmt: off
            self._company_client = (
                self._OrganisationSequence__company_client_class(  # type: ignore # noqa: E501
                    # *args,
                    # **kwargs,
                    **self.company_client_params,
                )
            )
            # fmt: on
            # self.__company_client = self.__company_client_class(
            # )
            # self._test_client = self.__company_client_class
        # assert False
        # return self._company_client.get_networks(*args, **kwargs)
        return self._company_client

    # @property
    # def company_client(self,
    #                    *args: QueryParameters,
    #                    **kwargs: QueryParameters) -> Graph:
    #     if not hasattr(self, "_company_client"):
    #         self._company_client = (
    #                 self._OrganisationSequence__company_client_class(
    #                     **self.company_client_params
    #                 ))
    #         # self.__company_client = self.__company_client_class(
    #         # )
    #         # self._test_client = self.__company_client_class
    #     # assert False
    #     return self._company_client.get_networks(*args, **kwargs)

    # @property
    def charity_client(
        self, *args: QueryParameters, **kwargs: QueryParameters
    ) -> Graph:
        """Cache and return results from get_charity_network method.

        Todo:
            * Refactor charity client to resemble company client class.
        """
        # pass parameters to function until refractor as client class
        # if not hasattr(self, "_charity_client"):
        #     self._charity_client = (
        #         self._OrganisationSequence__charity_client( *args, **kwargs,
        #         **self.charity_client_params))
        # return self._charity_client
        # return self._OrganisationSequence__charity_client(
        #        *args, **kwargs, **self.charity_client_params)
        return get_charity_network(
            *args,  # type: ignore[arg-type]
            **{
                **kwargs,  # type: ignore[arg-type]
                **self.charity_client_params,  # type: ignore[arg-type]
            },
        )

    def _get_charity_networks_generator(
        self,
        set_attr: bool = True,
        yield_attr: bool = False,
        # *args: QueryParameters,
        **kwargs: QueryParameters,
    ) -> Generator[Tuple[OrganisationEntry, Graph], None, None]:
        """Iterate organisations' ``charity_id`` and query their networks."""
        for organisation in self:
            if (
                not organisation._skip_charity
                and organisation.charity_id
                and (
                    not hasattr(organisation, "charity_network")
                    or organisation.charity_network is None
                )
            ):
                logger.info(
                    f"Querying charity {organisation.charity_id} "
                    f"for {organisation.name}..."
                )
                dict_args = {
                    **self.charity_client_params,
                    **kwargs,  # type: ignore[arg-type]
                }
                self._charity_runs.append(
                    CharityRunConfig(
                        root_id=organisation.charity_id,
                        start_time=datetime.now(),
                        end_time=None,
                        kinds_ids_dict=None,
                        # Params need apprasing for inclusion of locals
                        parameter_state=dict_args,
                        connected_components_count=None,
                        success=None,
                        composed_runs=[],
                    )
                )
                network: Optional[Graph] = self.charity_client(
                    # organisation.charity_id, *args, **dict_args)
                    organisation.charity_id,
                    **dict_args,  # type: ignore[arg-type]
                )
                self._charity_runs[-1].end_time = datetime.now()
                # self._charity_runs[-1]['kinds_ids_dict']
                if network:
                    self._charity_runs[-1].success = True
                    self._charity_runs[
                        -1
                    ].connected_components_count = number_connected_components(
                        network
                    )
                    self._charity_runs[-1].kinds_ids_dict = get_kinds_ids_dict(
                        network, CHARITY_NETWORK_KINDS
                    )
                    if set_attr:
                        setattr(organisation, "charity_network", network)
                else:
                    self._charity_runs[-1].success = False
                    self._charity_runs[-1].connected_components_count = None
            if yield_attr:
                if hasattr(organisation, "charity_network"):
                    yield organisation, organisation.charity_network
        self._charity_networks_cached = True

    def _get_company_networks_generator(
        self,
        set_attr: bool = True,
        yield_attr: bool = False,
        *args: QueryParameters,
        **kwargs: QueryParameters,
    ) -> Generator[Tuple[OrganisationEntry, Graph], None, None]:
        """Iterate organisations and query for company_network attribute.

        Todo:
            * Consider whether it's safer to return None for non company
              entries.
        """
        for organisation in self:
            if not organisation._skip_company and (
                not self._company_networks_cached
                and organisation.company_id
                and (
                    not hasattr(organisation, "company_network")
                    or organisation.company_network is None
                )
            ):
                logger.info(
                    f"Querying company {organisation.company_id} "
                    f"for {organisation.name}..."
                )
                # dict_args = {**self.company_client_params, **kwargs}
                # network: Optional[Graph] = self.company_client.get_network(
                #         organisation.company_id, *args, **dict_args)
                network: Optional[Graph] = self.company_client.get_network(
                    organisation.company_id, *args, **kwargs
                )
                self._company_runs.append(self._company_client._runs[-1])
                if set_attr:
                    setattr(organisation, "company_network", network)
            if yield_attr:
                if hasattr(organisation, "company_network"):
                    yield organisation, organisation.company_network
        self._company_networks_cached = True

    def get_company_networks_generator(
        self,
        reset: bool = False,
        *args: QueryParameters,
        **kwargs: QueryParameters,
    ) -> Generator[Tuple[OrganisationEntry, Graph], None, None]:
        """Yield each company graph.

        Todo:
            * Document the set_attr and yield_attr arguments
            * Consider ways to avoid confusion of charity graphs are
              independent but company graphs are not by default.
            * Consider taking advantage of just passing a list of company_ids
            * Add options for resetting/requirying
            * Add option for including all organisations or just those with
              companies
        """
        if reset:
            self._company_networks_cached = False
        for organisation, network in self._get_company_networks_generator(
            yield_attr=True,  # type: ignore[misc] # (multiple values passed)
            *args,  # type: ignore[arg-type]
            **kwargs,  # type: ignore[arg-type]
        ):
            yield organisation, network

        # orgs = [(organisation, deepcopy(organisation.company_network)) for
        #         organisation in self if hasattr(self, 'company_network')]
        # assert False
        # return orgs

    def get_charity_networks_generator(
        self,
        reset: bool = False,
        *args: QueryParameters,
        **kwargs: QueryParameters,
    ) -> Generator[Tuple[OrganisationEntry, Graph], None, None]:
        """Yield each charity graph.

        Todo:
            * Abstract to avoid copying between companies and charities
            * async
        """
        if reset:
            self._charity_networks_cached = False
        for organisation, network in self._get_charity_networks_generator(
            yield_attr=True,  # type: ignore[misc] # (multiple values passed)
            *args,  # type: ignore[arg-type]
            **kwargs,  # type: ignore[arg-type]
        ):
            yield organisation, network

    def get_charity_networks(self, *args, **kwargs) -> Sequence[Graph]:
        """Return list from get_charity_networks_generator."""
        return list(self.get_charity_networks_generator(*args, **kwargs))

    def get_company_networks(self, *args, **kwargs) -> Sequence[Graph]:
        """Return list from get_company_networks_generator."""
        return list(self.get_company_networks_generator(*args, **kwargs))

    def get_composed_company_network(
        self,
        # cache: bool = False,
        records: Optional[int] = None,
        # start_entry: int = 0,
        run_index: int = -1,
        *args: QueryParameters,
        **kwargs: QueryParameters,
    ) -> Graph:
        """Return a composed network of all in self.Organisations companies.

        Todo:
            * Take advantage of cache option to just compose networks
            * Add option to cache this attribute on class
            * Consider adding decorator for managing parameter_states
            * Add a run summary to _company_runs
            * Refactor to use self.company_ids
            * Ensure adding composed_runs doesn't aggregate extraneously
        """
        if self._company_networks_cached:
            if not self._company_composed_runs:
                self._company_composed_runs.append(
                    compose_all(
                        graph
                        for _, graph in self.get_company_networks(
                            *args, **kwargs
                        )
                    )
                )
            return self._company_composed_runs[run_index]
        # elif cache:
        #     logging.warning("Caching results during query so company_network"
        #                     "components are increasing in size in the order "
        #                     "they are queried, in this case starting with"
        #                     f"{self[0]} and ending with {self[-1]}.")
        else:
            initial_parameter_state: QueryParameters = (
                self.company_client._parameter_state
            )
            # Todo: assess whether to use self._company_runs instead
            prior_runs: int = len(self.company_client._runs)

            composed_graph: Graph = self.company_client.get_composed_network(
                root_ids=self.active_company_ids,
                parameter_states=(
                    {"compose_queried_networks": True, "_reset_cache": False}
                    for o in self.active_company_ids
                ),
                *args,
                **kwargs,
            )

            self._company_composed_runs.append(composed_graph)
            first_run: CompanyRunConfig = self.company_client._runs[
                prior_runs + 1
            ]
            last_run: CompanyRunConfig = self.company_client._runs[-1]
            # Todo: ensure this doesn't end up including unhelpful runs
            # Todo: check final ``connected_components_count``
            self._company_runs.append(
                CompanyRunConfig(
                    start_time=first_run.start_time,
                    end_time=last_run.end_time,
                    kinds_ids_dict=first_run.kinds_ids_dict,
                    parameter_state=first_run.parameter_state,
                    root_id=tuple(self.active_company_ids),
                    connected_components_count=(
                        last_run.connected_components_count
                    ),
                    composed_runs=[
                        r for r in self.company_client._runs[prior_runs:]
                    ],
                )
            )
            self.company_client._parameter_state = initial_parameter_state
            return composed_graph

    # @property
    def get_composed_charity_network(
        self,
        run_index: int = -1,
        *args: QueryParameters,
        # records: Optional[int] = None,
        # start_entry: int = 0,
        **kwargs: QueryParameters,
    ) -> Graph:
        """Iterate over charities, compose networks and return."""
        try:
            if not self._charity_networks_cached:
                prior_runs: int = len(self._charity_runs)
                composed_network: Graph = compose_all(
                    graph for _, graph in self.get_charity_networks()
                )
                self._charity_composed_runs.append(composed_network)
                first_run: CharityRunConfig = self._charity_runs[
                    prior_runs + 1
                ]
                last_run: CharityRunConfig = self._charity_runs[-1]
                self._charity_runs.append(
                    CharityRunConfig(
                        start_time=first_run.start_time,
                        end_time=last_run.end_time,
                        kinds_ids_dict=first_run.kinds_ids_dict,
                        parameter_state=first_run.parameter_state,
                        root_id=tuple(self.active_charity_ids),
                        connected_components_count=number_connected_components(
                            composed_network
                        ),
                        composed_runs=[
                            r for r in self._charity_runs[prior_runs:]
                        ],
                    )
                )
                return self._charity_composed_runs[-1]
            return self._charity_composed_runs[run_index]
            # return compose_all(n.charity_network for n in self
            #                    if hasattr(n, 'charity_network'))
        except StopIteration:
            logger.warning("No charity networks to compose.")
        # if not hasattr(self, '_charity_network'):
        #     self.get_charity_networks(*args, **kwargs)
        #     self._charity_network: Graph = compose_all(
        #             n for o, n in self if hasattr(o, 'charity_network'))
        # return deepcopy(self._charity_network)

    def get_networks(
        self,
        records: Optional[int] = None,
        # start_entry: int = 0,
        # branches: Optional[int] = None,
        # companies: bool = True,
        # charities: bool = True,
        composed: bool = True,
        correct_seed_graphs: bool = True,
        log_file: bool = True,
        *args: QueryParameters,
        **kwargs: QueryParameters,
    ) -> Tuple[Optional[Graph], Optional[Graph]]:
        """Query company and charity board networks and return networks."""
        # if logging_level:
        #     logger.setLevel(logging_level)
        # if log_path != LOG_PATH:
        #     # logging_level = logging_level or logging.DEBUG
        #     log_path = Path(log_path)
        #     log_path.parent.mkdir(exist_ok=True, parents=True)
        #     if reset_log:
        #         file_handler = logging.FileHandler(log_path, mode='w')
        #     else:
        #         file_handler = logging.FileHandler(log_path)
        #     file_handler.setLevel(logging_level)
        #     logger.addHandler(file_handler)
        # if not self.organisations_list:
        #     organisations_list = get_organisations_list(csv_path,
        #     csv_encoding)
        # records = records or len(organisations_list)
        if log_file:
            logger.addHandler(
                file_log_handler(
                    *args,  # type: ignore[arg-type]
                    **{**kwargs, **self.log_params},  # type: ignore[arg-type]
                )
            )
        start: datetime = datetime.now()
        logger.info(f"Start: {start.strftime(LOG_TIME_FORMAT)}")
        # if start_entry:
        #     logger.info(f'Beginning with record {start_entry}')
        try:
            if composed:
                charity_network: Graph = self.get_composed_charity_network(
                    *args, **kwargs  # type: ignore[arg-type]
                )
                company_network: Graph = self.get_composed_company_network(
                    *args, **kwargs  # type: ignore[arg-type]
                )

            # if correct_seed_graphs:
            #     charity_network = self.get_composed_company_network()
            #     company_network = self.get_company_networks()

            # self._companies_network = self.get_composed_company_network()
            # for i, organisation in enumerate(self[start_entry:records]):
            # if i % 10 is 0:
            # # Sort out printing timing estimates
            #     deciles = time.localtime()
            #     logger.info('Average time per organisation: {}'.format(
            #     ))
            # logger.info(f'{organisation.name:80} {i+1}/'
            #             f'{records - start_entry}')
            # organisation._get_networks(branches)
            # print()
        finally:
            end: datetime = datetime.now()
            logger.info(f"End: {end.strftime(LOG_TIME_FORMAT)}")
            # Note: timedelta doesn't support strftime, probably remove '()'
            logger.info(f"Total Time: {(end - start)}")
        # if pickle_path:
        #     with open(pickle_path, 'wb') as pickle_file:
        #         pickle.dump(organisations_list, pickle_file)
        # return self._company_network, self._charity_network
        return charity_network, company_network

    def _write_network(
        self,
        network_type: str,
        path: PathLike = JSON_DATA_PATH,
        run_index: int = -1,  # Default is most recent
    ) -> None:
        graph: Graph = getattr(self, f"_{network_type}_composed_runs")[
            run_index
        ]
        config_dict: Union[CharityRunConfig, CompanyRunConfig] = [
            r
            for r in getattr(self, f"_{network_type}_runs")
            if "composed_runs" in r
        ][run_index]
        json_filename: str = get_network_json_file_name(network_type)
        path = Path(path) / json_filename
        # Todo: Coerce config_dict type to fit Optional[Dict[str, Any]]
        write_json_graph(graph, path, config_dict)  # type: ignore[arg-type]

    def write_networks(
        self,
        composed: bool = True,
        charities: bool = True,
        companies: bool = True,
        path: PathLike = JSON_DATA_PATH,
    ) -> None:
        """Write json files of saved company and/or charity networks."""
        if composed:
            if charities:
                self._write_network("charity", path)
            if companies:
                self._write_network("company", path)

    def _read_network(
        self,
        network_type: str,
        path: PathLike = JSON_DATA_PATH,
        add_meta_data: bool = True,
        latest: bool = False,
        store: bool = True,
    ) -> Graph:
        if latest:
            path = get_latest_json_file_name(network_type, path)
        if add_meta_data:
            graph, metadata = read_json_graph(path, True)
            getattr(self, f"_{network_type}_runs").append(metadata)
        else:
            graph = read_json_graph(path)
        if store:
            getattr(self, f"_{network_type}_composed_runs").append(graph)
        return graph

    def _read_network_wrapper(
        self,
        network_type: str,
        path: PathLike = JSON_DATA_PATH,
        add_meta_data: bool = True,
        populate: bool = False,
        latest: bool = False,
        store: bool = False,
        *args,
        **kwargs,
    ) -> Tuple[Graph, Optional[Sequence[OrganisationNetworkDict]]]:
        """Wrap network genrator and storage."""
        networks_sequence: Optional[Sequence[OrganisationNetworkDict]] = None
        loaded_graph: Graph = self._read_network(
            network_type, path, add_meta_data, latest, store
        )
        if populate:
            networks_sequence = [
                OrganisationNetworkDict(
                    network=graph,
                    organisation=org,
                    organisation_id=org_id,
                    organisation_type=network_type,
                )
                # {"network": graph, "organisation": org,
                # "organisation_id": c_id}
                for graph, org_id, org in self._ego_networks_generator(
                    network_type, *args, **kwargs
                )
            ]
        return loaded_graph, networks_sequence
        Tuple[Graph, Union[CharityIDType, CompanyIDType], OrganisationEntry],
        None,
        None,

    def read_charity_network(
        self, **kwargs
    ) -> Tuple[Graph, Optional[Sequence[OrganisationNetworkDict]]]:
        """Read charity network from path and store in instance."""
        return self._read_network(network_type="charity", **kwargs)

    def read_company_network(
        self, **kwargs
    ) -> Tuple[Graph, Optional[Sequence[OrganisationNetworkDict]]]:
        """Read company network from path and store in instance."""
        return self._read_network(network_type="company", **kwargs)

    # def read_company_network(self,
    #                          composed: bool = True,
    #                          path: PathLike = JSON_DATA_PATH,
    #                          add_meta_data: bool = True,
    #                          populate: bool = False,
    #                          latest: bool = True, *args, **kwargs):
    #     """Read company network from path and store in instance."""
    #     networks: OrganisationNetworkSequence = []
    #     self._read_network("company", composed, path, add_meta_data, latest)
    #     if populate:
    #         networks = [
    #                 {'network': graph, 'organisation': org, 'company_id':
    #                 c_id} for graph, org, c_id in
    #                 self._ego_networks_gernator('company', *args, **kwargs)]
    #         if composed:
    #             compose_all(g for g, org, c_id in network)

    def read_networks(
        self,
        charities: bool = True,
        companies: bool = True,
        populate: bool = False,
        *args,
        **kwargs,
    ) -> Optional[Tuple[Graph, Graph]]:
        """Read saved graph files of charities and/or companies."""
        charity_network: Optional[Graph] = None
        company_network: Optional[Graph] = None
        if charities:
            charity_network = self.read_charity_network(*args, **kwargs)
        if companies:
            company_network = self.read_company_network(*args, **kwargs)
        return charity_network, company_network

    def get_charity_ego_networks(self, *args, **kwargs) -> Graph:
        """Return charity networks iterator seeded from self.organisations."""
        return list(self._ego_networks_generator("charity", *args, **kwargs))

    def get_company_ego_networks(self, *args, **kwargs) -> Graph:
        """Return company networks iterator seeded from self.organisations."""
        return list(self._ego_networks_generator("company", *args, **kwargs))

    def get_composed_charity_ego_network(self, *args, **kwargs) -> Graph:
        """Return composed charity network seeded from self.organisations."""
        return compose_all(
            graph
            for graph, org, network_id in self._ego_networks_generator(
                "charity", *args, **kwargs
            )
        )

    def get_composed_company_ego_network(self, *args, **kwargs) -> Graph:
        """Return composed company network seeded from self.organisations."""
        return compose_all(
            graph
            for graph, org, network_id in self._ego_networks_generator(
                "company", *args, **kwargs
            )
        )

    def _ego_networks_generator(
        self,
        network_type: str,
        store: bool = True,
        hops: int = 0,
        radius: Optional[int] = None,
        from_composed: bool = False,
        # compose: bool = True,
        run_index: int = -1,
    ) -> Generator[
        Tuple[Graph, Union[CharityIDType, CompanyIDType], OrganisationEntry],
        None,
        None,
    ]:
        """Generate local ego networks per org from active nodes.

        Note:
            Radius takes precedent over hops.

        Todo:
            * Check conversion of hops to radius
            * Raise errors if radius is too big or negative
            * Consider breaking into multiple functions
            * Add logging info
            * Add caching
        """
        graph: Graph
        radius = radius or hops * 2 + 1
        try:
            if from_composed:
                graph = getattr(self, f"_{network_type}_composed_runs")[
                    run_index
                ]
            else:
                graph = getattr(self, f"_{network_type}_runs")[run_index]
        except IndexError:
            raise NoLoadedNetworkDataError

        assert len(graph) > 0

        # if compose:
        #     graph_composed: Graph = Graph()
        for organisation in self:
            if getattr(organisation, f"{network_type}_id") and not getattr(
                organisation, f"_skip_{network_type}"
            ):
                try:
                    org_graph: Graph = ego_graph(
                        graph,
                        getattr(organisation, f"{network_type}_id"),
                        radius,
                    )
                except NodeNotFound:
                    logger.error(
                        f"{organisation.name} {network_type} not "
                        f"found. Failed to add an ego network of "
                        f"{hops} hops ({radius} radius)."
                    )
                    continue
                if store:  # Todo: indicate number of hops in network attr name
                    setattr(organisation, f"{network_type}_network", org_graph)
                yield org_graph, getattr(
                    organisation, f"{network_type}_id"
                ), organisation

    # def _populate_ego_networks(self,
    #                            radius: int = 1,
    #                            company: bool = True,
    #                            charity: bool = True,
    #                            compose: bool = True,
    #                            run_index: int = -1,
    #                            ) -> Optional[Tuple[Graph, Graph]]:
    #     """Generate local ego networks per org from active nodes.

    #     Todo:
    #         * Raise errors if radius is too big or negative
    #         * Consider breaking into multiple functions
    #         * Add logging info
    #     """
    #     if charity:
    #         charity_graph: Graph = self._charity_composed_runs[run_index]
    #         if compose:
    #             charity_graph_composed: Graph = Graph()
    #     if company:
    #         company_graph: Graph = self._company_composed_runs[run_index]
    #         if compose:
    #             company_graph_composed: Graph = Graph()
    #     for organisation in self:
    #         if organisation.charity_id and not organisation._skip_charity:
    #             org_charity_graph: Graph = ego_graph(charity_graph,
    #                                                  organisation.charity_id,
    #                                                  radius)
    #             setattr(organisation, 'charity_network', org_charity_graph)
    #             if compose:
    #                 charity_graph_composed = compose(charity_graph_composed,
    #                                                  org_charity_graph)
    #         if organisation.company_id and not organisation._skip_company:
    #             org_company_graph: Graph = ego_graph(company_graph,
    #                                                  organisation.company_id,
    #                                                  radius)
    #             setattr(organisation, 'company_network', org_company_graph)
    #             if compose:
    #                 company_graph_composed = compose(company_graph_composed,
    #                                                  org_company_graph)
    #     if compose:
    #         return charity_graph, company_graph



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `ukboards` command line interface."""

import inspect
from typing import Final

import pytest
from click.testing import CliRunner

from ukboards.cli import ukboards
from ukboards.companies import COMPANIES_HOUSE_URL, CompanyIDType

PUNCHDRUNK_COMPANY_ID: Final[CompanyIDType] = "04547069"
PUNCHDRUNK_JSON_SUBSET: Final[
    str
] = '{"status": "active", "company_name": "PUNCHDRUNK"}'

CORRECT_HELP = """\
Usage: ukboards [OPTIONS] COMMAND [ARGS]...

  Query UK company and charity board data.

Options:
  -a, --api-keys-path FILE  Path to file with API keys (default=.env)
  -i, --indent INTEGER      How many spaces to indent printing json queries.
                            [default: 2]
  --help                    Show this message and exit.

Commands:
  charity            Query Charities Commision by charity id.
  company            Query Companies House by company id.
  csv-organisations  Path to csv with company and charity ids.
"""


@pytest.fixture
def cli_runner() -> CliRunner:
    """Fixture for adding CliRunners() to tests."""
    return CliRunner()


def convert_output_string(output: str) -> str:
    """Fix formatting of string to conform with other Companise House tests."""
    return output.strip('\n"').replace('\\"', '"')


class TestMainCommandLineInterface:
    """Test ukboards root command line interface.

    Todo:
        * Add options for processing CSV files from the command line.
    """

    def test_command_line_interface(self, cli_runner):
        """Test the CLI."""
        result = cli_runner.invoke(ukboards)
        assert result.exit_code == 0
        assert CORRECT_HELP in result.output
        help_result = cli_runner.invoke(ukboards, ["--help"])
        assert help_result.exit_code == 0
        assert CORRECT_HELP in help_result.output

    # def test_csv_organisations(self):
    #     """Test loading a list of organisations from a csv file."""
    #     pass


class TestCompaniesCommandLineInterface:
    """Test Companies options."""

    COMPANIES_HELP: Final[
        str
    ] = """\
    Usage: ukboards company [OPTIONS] COMPANY_ID

      Query Companies House by company id.

    Options:
      -k, --api-key TEXT
      --help              Show this message and exit."""

    def test_companies_help(self, cli_runner):
        """Test help output for companies subcommand."""
        result = cli_runner.invoke(ukboards, ["company", "--help"])
        assert result.exit_code == 0
        assert inspect.cleandoc(self.COMPANIES_HELP) in result.output

    def test_default_query(self, requests_mock, caplog, cli_runner):
        """Test mock querying PUNCHDRUNK."""
        requests_mock.get(
            f"{COMPANIES_HOUSE_URL}/company/" f"{PUNCHDRUNK_COMPANY_ID}",
            json=PUNCHDRUNK_JSON_SUBSET,
        )
        result = cli_runner.invoke(
            ukboards, ["company", PUNCHDRUNK_COMPANY_ID]
        )
        assert convert_output_string(result.output) == PUNCHDRUNK_JSON_SUBSET
        assert caplog.records == []

    # def test_passing_env_path(self, requests_mock, caplog):
    #     """Test mock querying PUNCHDRUNK."""
    #     runner = CliRunner()


class TestCharitiesCommandLineInterface:
    """Test Charity command line options."""

    CHARITIES_HELP: str = """\
    Usage: ukboards charity [OPTIONS] CHARITY_ID

      Query Charities Commision by charity id.

    Options:
      -k, --api-key TEXT
      --help              Show this message and exit."""
    PHOTOGRAPHERS_GALLERY_NAME: str = (
        "THE PHOTOGRAPHERS' GALLERY LTD                    "
        "                                                  "
        "                                                  "
    )
    PHOTOGRAPHERS_GALLERY_NUMBER: int = 262548

    def test_charities_help(self, cli_runner):
        """Test help output for companies subcommand."""
        result = cli_runner.invoke(ukboards, ["charity", "--help"])
        assert result.exit_code == 0
        assert inspect.cleandoc(self.CHARITIES_HELP) in result.output

    @pytest.mark.remote_data
    def test_charities_query(self, cli_runner):
        """Test basic charity query."""
        result = cli_runner.invoke(
            ukboards, ["charity", str(self.PHOTOGRAPHERS_GALLERY_NUMBER)]
        )
        assert result.exit_code == 0
        assert (
            f"'CharityNumber': {self.PHOTOGRAPHERS_GALLERY_NUMBER},"
            in result.output
        )
        assert (
            f"'CharityName': \"{self.PHOTOGRAPHERS_GALLERY_NAME}\","
            in result.output
        )
